#%%
import QuantLib as ql
from datetime import datetime
import pandas as pd
import time
import py_vollib_vectorized 
from pyvol_terminal.instruments import instruments
from pyvol_terminal.instruments import builders as builders_instruments
from pyvol_terminal import utils as main_utils
from pyvol_terminal.instruments import instruments
from pyvol_terminal.instruments import builders as builders_instruments
from pyvol_terminal.engines.option_engines import black
from pyvol_terminal.engines import surface_engines
from pyvol_terminal.interfaces.volatility_surface.interface import Interface as interface_surface
#from pyvol_terminal.interfaces.volatility_table.interface import Interface as interface_vol_table
from pyvol_terminal.interfaces.option_monitor.interface import Interface as interface_omon_table
from pyvol_terminal.interfaces.pricing_engine.interface import Interface as interface_pricing_engine
#from pyvol_terminal.interfaces.slice.interface import Interface as interface_slice
from pyvol_terminal import misc_widgets
from pyvol_terminal.ui_themes import legend_colourmap_getter
from pyvol_terminal.data_classes import builders as builders_dataclasses
from examples.synthetic import websocket_streamer
import sys
import numpy as np
from PySide6 import QtGui   


def format_date(dt: datetime) -> str:
    return f"{dt.day}{dt.strftime('%b%y').upper()}"


def reformat_instrument_names(df, fut=False):
    expiries = pd.to_datetime(df['expiry'], utc=True, unit="s")
    instrument_names = []
    channels=[]
    for idx, expiry in enumerate(expiries):
        exp_str = format_date(expiry)
        tupl = df["instrument_name"].iloc[idx].split("-")
        if len(tupl) == 4:
            underlying, _, strike, flag = tupl
            instrument_name = f"{underlying}-{exp_str}-{strike}-{flag}"
        elif len(tupl) == 2:
            underlying, _ = tupl
            instrument_name = f"{underlying}-{exp_str}"
        channel = f"ticker.{instrument_name}.100ms"
        
        instrument_names.append(instrument_name)
        channels.append(channel)    
    df["instrument_name"] = instrument_names
    return channels
    

def get_saved_data():
    df_options = pd.read_csv("df_options.csv")
    df_futures = pd.read_csv("df_futures.csv")
    df_spot = pd.read_csv("df_spot.csv")
    
    reference_time = 1743506630.313821
    time_delta = time.time() - reference_time
    df_options["expiry"]+=time_delta
    df_futures["expiry"]+=time_delta

    channels_opt = reformat_instrument_names(df_options)
    channels_fut = reformat_instrument_names(df_futures)
    

    """
    expiry = df_options["expiry"].unique()[:6]
    df_options = df_options[df_options["expiry"].isin(expiry)]
    df_options = df_options[(df_options["strike"] <= 100_000) & (df_options["strike"] >= 70_000)]
    df_futures = df_futures[df_futures["expiry"].isin(expiry)]
    """
    
    with open("channels.txt", "r") as file:
        channels = [line.strip() for line in file]  
    
    channels_spot = [channel for channel in channels if "BTC_USDC" in channel]
    channels = channels_spot + channels_fut + channels_spot
        
    return channels, df_options, df_futures, df_spot


def get_data():
    
    channels, df_options, df_futures, df_spot = get_saved_data()
    
    vol_params = {"ATM_vol" : 0.65,
                  "kurtosis_1" : 0.95,
                  "kurtosis_2" : 0.2,
                  "decay" : 0.8
                  }

    btc_usdc = 90000.
    r = 0.05 
    
    df_options = df_options[df_options["flag"]=="p"]
    df_options = df_options[df_options["strike"] <= btc_usdc-2000]       
    print(df_options)
    df_options["underlying_px"] = np.nan
    
    df_options["bid_iv"] = np.nan
    df_options["ask_iv"] = np.nan
    df_options["best_bid_price"] = np.nan
    df_options["best_ask_price"] = np.nan
    
    df_options["expiry_T"] = (df_options["expiry"] - time.time()) / 3600 / 24 / 365
    df_options = df_options[df_options["expiry_T"] >0.05]       
    
    generate_futures(df_futures, btc_usdc, r)
    df_futures = df_futures[df_futures["expiry_T"] >0.05]     
    
    print(df_futures)
    
    
    for idx, row in df_options.iterrows():
        exp = row["expiry"]
        price = df_futures[df_futures["expiry"] == exp]["mid"].item()
        df_options.loc[idx, "underlying_px"] = price
        df_options.loc[idx, "underlying_ticker"] = df_futures[df_futures["expiry"] == exp]["instrument_name"].item()
        
    for idx, row in df_options.iterrows():
        price = row["underlying_px"]
        bid_iv = generate_synthetic_surface(vol_params, row["underlying_px"], df_options.loc[idx, "strike"], df_options.loc[idx, "expiry_T"]) + 0.01
        mid_iv = generate_synthetic_surface(vol_params, row["underlying_px"], df_options.loc[idx, "strike"], df_options.loc[idx, "expiry_T"]) 
        ask_iv = generate_synthetic_surface(vol_params, row["underlying_px"], df_options.loc[idx, "strike"], df_options.loc[idx, "expiry_T"]) - 0.01
        df_options.loc[idx, "mid_iv"] = mid_iv

        df_options.loc[idx, "bid_iv"] = bid_iv
        df_options.loc[idx, "ask_iv"] = ask_iv
        df_options.loc[idx, "best_bid_price"] = py_vollib_vectorized.vectorized_black(row["flag"], price, row["strike"], row["expiry_T"], r, bid_iv)["Price"].item() / price
        df_options.loc[idx, "best_ask_price"] = py_vollib_vectorized.vectorized_black(row["flag"], price, row["strike"], row["expiry_T"], r, ask_iv)["Price"].item() / price
        
    df_spot["bids"] = btc_usdc - 0.01
    df_spot["asks"] = btc_usdc + 0.01
    
    future_underlying_ticker_map = {future_name : df_spot["instrument_name"].values[0] for future_name in df_futures["instrument_name"]}
    
    option_underlying_ticker_map = {}
    for _, option_row in df_options.iterrows():
        option_name = option_row["instrument_name"]
        option_expiry = option_row["expiry"]
        future_name = df_futures[df_futures["expiry"] == option_expiry]["instrument_name"].iloc[0]
        option_underlying_ticker_map[option_name] = future_name
    return channels, df_options, df_futures, df_spot, option_underlying_ticker_map, future_underlying_ticker_map

def generate_futures(df_futures, spot, r):
    df_futures["expiry_T"] = (df_futures["expiry"] - time.time()) / 3600 / 24 / 365
    
    df_futures["mid"] = spot * np.exp(r * df_futures["expiry_T"])
    df_futures["mid"] = np.round(df_futures["mid"])
    df_futures["best_bid_price"] = (spot -0.5) * np.exp(r * df_futures["expiry_T"])
    df_futures["best_ask_price"] = (spot + 0.5) * np.exp(r * df_futures["expiry_T"])
    df_futures["r"] = r
    
def generate_synthetic_surface(market_params, S, K, T):
    ATM_vol = market_params["ATM_vol"]
    kurtosis_1 = market_params["kurtosis_1"]
    kurtosis_2 = market_params["kurtosis_2"]
    decay = market_params["decay"]
    
    if K < S:
        smile=ATM_vol + kurtosis_1 * ((K - S) / S) ** 2
    else:
        smile=ATM_vol + kurtosis_2 * ((K - S) / S) ** 2
    
    flattening_factor = np.exp(-decay * T)
    IV = smile * flattening_factor
    return IV

channels, df_options, df_futures, df_spot, option_underlying_ticker_map, future_underlying_ticker_map = get_data()


price_types = ["mid", "bid", "ask"]
class DummyEngine:
    def evaluate(self, x):
        return 0

interest_rate_config = {"engine" : DummyEngine(),
                        "use_ws_response" : False} 
dividend_rate_config = {"engine" : DummyEngine(),
                        "use_ws_response" : False} 

spot_config = {"object" : instruments.Spot,
                "data" : df_spot,
                "price_types" : price_types,
                "valid_price_checker" : main_utils.ValidPriceChecker(50)}

futures_config = {"object" : instruments.Future,
                    "data" : df_futures,
                    "price_types" : price_types,
                    "underlying_map" : future_underlying_ticker_map,
                    "underlying_px_type" : "mid", #{p : "mid" for p in ["bid", "ask", "mid"]},
                    "valid_price_checker" : main_utils.ValidPriceChecker(50),
                    "interest_rate_config" : interest_rate_config,
                    "dividend_rate_config" : dividend_rate_config
                    }

options_config = {"object" : instruments.OptionInverted,
                    "data" : df_options,
                    "price_types" : price_types,
                    "underlying_map": option_underlying_ticker_map,
                    "underyling_price_type" : "mid", #{p : "mid" for p in ["bid", "ask", "mid"]},
                    "engine": black.Engine,
                    "valid_price_checker" : main_utils.ValidPriceChecker(50),
                    "interest_rate_config" : interest_rate_config,
                    "dividend_rate_config" : dividend_rate_config
                    }




instruments_config = {"options" : options_config,
                        "futures" : futures_config,
                        "spot"    : spot_config,
                        }
instrument_manager = builders_instruments.create_instrument_manager(instruments_config)


#%%
spot = df_spot[[]]

import numpy as np

expiration_dt = [datetime(2025, 9,10), datetime(2025, 10,10), datetime(2025, 12,10), datetime(2026, 2,10)]

expiration_dates_ql = [ql.Date(dt.day, dt.month, dt.year) for dt in expiration_dt]

calculation_date = ql.Date(datetime.now().day, datetime.now().month, datetime.now().year)
        
v0, kappa, theta, rho, sigma = 0.01, 0.2, 0.02, -0.75, 0.5
dfs=[1.]
futures_exp_dates = [calculation_date]
spot_object = list(instrument_manager.spot_instrument_container.objects.values())[0]
spot = spot_object.mid



for futures_object in instrument_manager.futures_instrument_container.objects.values():
    exp_ts = futures_object.expiry
    exp_dt = datetime.fromtimestamp(exp_ts)
    exp_ql = ql.Date(exp_dt.day, exp_dt.month, exp_dt.year)
    dfs.append(futures_object.mid / spot)
    futures_exp_dates.append(exp_ql)
builders_data_classes.create_volatility_data_from_vol_vect(vol_vect_container, default_nTicks)
data_container = data_container["mid"]
xyz = np.column_stack(data_container.vol_vector.data())

def getRectStrikesExpiries(xyz) -> Tuple[np.ndarray, ...]:
    print(len(np.unique(xyz[:, 0])))
    strikes, expiries, vol = xyz[:,0], xyz[:,1], xyz[:,2]
    exp = np.unique(expiries)
    unique, counts = np.unique(strikes, return_counts=True)
    k = unique[counts >= exp.size]
    
    xyz = xyz[np.isin(xyz[:, 0], k)]
    xyz = xyz[np.isin(xyz[:, 1], exp)]
    
    return xyz


xyz = getRectStrikesExpiries(xyz)

strikes = np.unique(xyz[:, 0])
expiries = np.unique(xyz[:, 1])
vol_grid = xyz[:, 2].reshape(len(strikes), len(expiries))
print(vol_grid)
expiration_dt = [datetime.fromtimestamp(ts) for ts in expiries]
expiration_dates_ql = [ql.Date(dt.day, dt.month, dt.year) for dt in expiration_dt]



implied_vols = ql.Matrix(vol_grid.shape[0], vol_grid.shape[1])


for i, _ in enumerate(strikes):
    for j, _ in enumerate(expiries):
        implied_vols[i][j] = vol_grid[i][j]


calendar = ql.UnitedStates(ql.UnitedStates.NYSE)
day_count = ql.Actual365Fixed()
black_var_surface = ql.BlackVarianceSurface(calculation_date,
                                            calendar,
                                            expiration_dates_ql,
                                            strikes.tolist(),
                                            implied_vols,
                                            day_count
                                            )


ql.Settings.instance().evaluationDate = calculation_date

# Ensure you explicitly add the reference date with discount 1.0
futures_exp_dates = [calculation_date] + futures_exp_dates
dfs = [1.0] + dfs  # Ensure first discount = 1.0


calendar = ql.NullCalendar()
curve = ql.DiscountCurve(futures_exp_dates, dfs, ql.Actual360())
interest_ts = ql.YieldTermStructureHandle(curve)

dividend_ts = ql.YieldTermStructureHandle(ql.FlatForward(calculation_date, ql.QuoteHandle(ql.SimpleQuote(0.0)), ql.Actual365Fixed())),  # No dividend yield
process = ql.HestonProcess(interest_ts, dividend_ts,
                           ql.QuoteHandle(ql.SimpleQuote(spot)),
                           v0, kappa, theta, sigma, rho)

model = ql.HestonModel(process)
engine = ql.AnalyticHestonEngine(model)

#%%


def generate_synthetic_surface(market_params, S, k, t):
    ATM_vol = market_params["ATM_vol"]
    kurtosis_1 = market_params["kurtosis_1"]
    kurtosis_2 = market_params["kurtosis_2"]
    decay = market_params["decay"]
    vols=[]
    
    for Ki, T in zip(k,t):
        vol_i=[]
        for K in Ki:
            if K < S:
                smile=ATM_vol + kurtosis_1 * ((K - S) / S) ** 2
            else:
                smile=ATM_vol + kurtosis_2 * ((K - S) / S) ** 2
            
            flattening_factor = np.exp(-decay * T)
            IV = smile * flattening_factor
            vol_i.append(IV) 
        vols.append(vol_i)
    return vols


atm = 120
expiries = 4
max_strikes = 9  # must be > 8 as per your requirement

# Decreasing number of strikes per expiry
strikes_per_expiry = [max_strikes - i for i in range(expiries)]  # [9, 8, 7, 6]

strike_step = 2  # spacing between strikes
strike_lists = []

for n in strikes_per_expiry:
    half = n // 2
    # Center around ATM, offset evenly
    if n % 2 == 1:
        strikes = [atm + (i - half) * strike_step for i in range(n)]
    else:
        strikes = [atm + (i - half + 0.5) * strike_step for i in range(n)]
    # Only include strikes > 8
    strikes = [s for s in strikes if s > 8]
    strike_lists.append(strikes)

# Result
for i, expiry_strikes in enumerate(strike_lists, 1):
    print(f"Expiry {i}: {expiry_strikes}")

vol_params = {"ATM_vol" : 0.65,
              "kurtosis_1" : 1.05,
              "kurtosis_2" : 0.35,
              "decay" : 0.3
              }

vols = generate_synthetic_surface(vol_params, spot, strike_lists, dt_y)


#%%
import QuantLib as ql
import math


import numpy as np
import time
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
plt.rcParams['figure.figsize']=(15,7)
plt.style.use("dark_background")
from matplotlib import cm
import gc


total_seconds = 0

day_count = ql.Actual365Fixed()
calendar = ql.UnitedStates(ql.UnitedStates.NYSE)

calculation_date = ql.Date(6, 11, 2015)

spot = 659.37
ql.Settings.instance().evaluationDate = calculation_date

dividend_yield = ql.QuoteHandle(ql.SimpleQuote(0.0))
risk_free_rate = 0.01
dividend_rate = 0.0
flat_ts = ql.YieldTermStructureHandle(
    ql.FlatForward(calculation_date, risk_free_rate, day_count))
dividend_ts = ql.YieldTermStructureHandle(
    ql.FlatForward(calculation_date, dividend_rate, day_count))



"""
expiration_dates = [ql.Date(9,12,2021), ql.Date(9,1,2022), ql.Date(9,2,2022),
                    ql.Date(9,3,2022), ql.Date(9,4,2022), ql.Date(9,5,2022),
                    ql.Date(9,6,2022), ql.Date(9,7,2022), ql.Date(9,8,2022),
                    ql.Date(9,9,2022), ql.Date(9,10,2022), ql.Date(9,11,2022),
                    ql.Date(9,12,2022), ql.Date(9,1,2023), ql.Date(9,2,2023),
                    ql.Date(9,3,2023), ql.Date(9,4,2023), ql.Date(9,5,2023),
                    ql.Date(9,6,2023), ql.Date(9,7,2023), ql.Date(9,8,2023),
                    ql.Date(9,9,2023), ql.Date(9,10,2023), ql.Date(9,11,2023)]


strikes_data = {
    ql.Date(9,12,2021): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34],
    ql.Date(9,1,2022): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34, 725.31],
    ql.Date(9,2,2022): [560.46, 593.43, 626.40, 659.37, 692.34, 725.31],
    ql.Date(9,3,2022): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34, 725.31, 758.28],
    ql.Date(9,4,2022): [560.46, 593.43, 626.40, 659.37, 692.34, 725.31],
    ql.Date(9,5,2022): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34, 725.31],
    ql.Date(9,6,2022): [593.43, 626.40, 659.37, 692.34, 725.31],
    ql.Date(9,7,2022): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34, 725.31, 758.28],
    ql.Date(9,8,2022): [560.46, 593.43, 626.40, 659.37, 692.34],
    ql.Date(9,9,2022): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34, 725.31],
    ql.Date(9,10,2022): [593.43, 626.40, 659.37, 692.34, 725.31, 758.28],
    ql.Date(9,11,2022): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34, 725.31],
    ql.Date(9,12,2022): [560.46, 593.43, 626.40, 659.37, 692.34, 725.31],
    ql.Date(9,1,2023): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34, 725.31, 758.28],
    ql.Date(9,2,2023): [560.46, 593.43, 626.40, 659.37, 692.34],
    ql.Date(9,3,2023): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34, 725.31],
    ql.Date(9,4,2023): [593.43, 626.40, 659.37, 692.34, 725.31, 758.28],
    ql.Date(9,5,2023): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34],
    ql.Date(9,6,2023): [560.46, 593.43, 626.40, 659.37, 692.34, 725.31],
    ql.Date(9,7,2023): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34, 725.31, 758.28],
    ql.Date(9,8,2023): [560.46, 593.43, 626.40, 659.37, 692.34],
    ql.Date(9,9,2023): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34, 725.31],
    ql.Date(9,10,2023): [593.43, 626.40, 659.37, 692.34, 725.31, 758.28],
    ql.Date(9,11,2023): [527.50, 560.46, 593.43, 626.40, 659.37, 692.34]
}


vol_data = {
    ql.Date(9,12,2021): [0.37819, 0.34177, 0.30394, 0.27832, 0.26453, 0.25916],
    ql.Date(9,1,2022): [0.3445, 0.31769, 0.2933, 0.27614, 0.26575, 0.25729, 0.25228],
    ql.Date(9,2,2022): [0.37419, 0.35372, 0.33729, 0.32492, 0.31601, 0.30883],
    ql.Date(9,3,2022): [0.37498, 0.35847, 0.34475, 0.33399, 0.32715, 0.31943, 0.31098, 0.30506],
    ql.Date(9,4,2022): [0.35941, 0.34516, 0.33296, 0.32275, 0.31867, 0.30969],
    ql.Date(9,5,2022): [0.35521, 0.34242, 0.33154, 0.3219, 0.31948, 0.31096, 0.30424],
    ql.Date(9,6,2022): [0.35442, 0.34267, 0.33288, 0.32374, 0.32245],
    ql.Date(9,7,2022): [0.35384, 0.34286, 0.33386, 0.32507, 0.3246, 0.31745, 0.31135, 0.306],
    ql.Date(9,8,2022): [0.35338, 0.343, 0.33464, 0.32614, 0.3263],
    ql.Date(9,9,2022): [0.35301, 0.34312, 0.33526, 0.32698, 0.32766, 0.32132, 0.31558],
    ql.Date(9,10,2022): [0.35272, 0.34322, 0.33574, 0.32765, 0.32873, 0.32267],
    ql.Date(9,11,2022): [0.35246, 0.3433, 0.33617, 0.32822, 0.32965, 0.32383, 0.31831],
    ql.Date(9,12,2022): [0.35226, 0.34336, 0.33651, 0.32869, 0.3304, 0.32477],
    ql.Date(9,1,2023): [0.35207, 0.34342, 0.33681, 0.32911, 0.33106, 0.32561, 0.32025, 0.3155],
    ql.Date(9,2,2023): [0.35171, 0.34327, 0.33679, 0.32931, 0.3319],
    ql.Date(9,3,2023): [0.35128, 0.343, 0.33658, 0.32937, 0.33276, 0.32769, 0.32255],
    ql.Date(9,4,2023): [0.35086, 0.34274, 0.33637, 0.32943, 0.3336, 0.32872],
    ql.Date(9,5,2023): [0.35049, 0.34252, 0.33618, 0.32948, 0.33432, 0.32959],
    ql.Date(9,6,2023): [0.35016, 0.34231, 0.33602, 0.32953, 0.33498, 0.3304],
    ql.Date(9,7,2023): [0.34986, 0.34213, 0.33587, 0.32957, 0.33556, 0.3311, 0.32631, 0.32217],
    ql.Date(9,8,2023): [0.34959, 0.34196, 0.33573, 0.32961, 0.3361],
    ql.Date(9,9,2023): [0.34934, 0.34181, 0.33561, 0.32964, 0.33658, 0.33235, 0.32769],
    ql.Date(9,10,2023): [0.34912, 0.34167, 0.3355, 0.32967, 0.33701, 0.33288],
    ql.Date(9,11,2023): [0.34891, 0.34154, 0.33539, 0.3297, 0.33742, 0.33337]
}


all_strikes = set()
for strikes in strikes_data.values():
    all_strikes.update(strikes)
all_strikes = sorted(all_strikes)
n=10000
for i in range(n):

    s = time.time()
    implied_vols = ql.Matrix(len(all_strikes), len(expiration_dates))
    for j, date in enumerate(expiration_dates):
        date_strikes = strikes_data[date]
        date_vols = vol_data[date]
        for i, strike in enumerate(all_strikes):
            if strike in date_strikes:
                idx = date_strikes.index(strike)
                implied_vols[i][j] = date_vols[idx]
            else:
                implied_vols[i][j] = np.nan  # or use ql.Null<Real>()

    black_var_surface = ql.BlackVarianceSurface(
        calculation_date, calendar,
        expiration_dates, all_strikes,
        implied_vols, day_count)

    black_var_surface.enableExtrapolation()

    strikes_grid = strikes_grid = np.linspace(min(all_strikes)*0.8, max(all_strikes)*1.2, 100)
    expiry = 1.0 

    implied_vols = [black_var_surface.blackVol(expiry, s)for s in strikes_grid]
    total_seconds+=time.time() - s
print(f"total_seconds: {total_seconds}")
print(f"per sec: {n/total_seconds}")
"""
expiration_dates = [ql.Date(6,12,2015), ql.Date(6,1,2016), ql.Date(6,2,2016),
                    ql.Date(6,3,2016), ql.Date(6,4,2016), ql.Date(6,5,2016), 
                    ql.Date(6,6,2016), ql.Date(6,7,2016), ql.Date(6,8,2016),
                    ql.Date(6,9,2016), ql.Date(6,10,2016), ql.Date(6,11,2016), 
                    ql.Date(6,12,2016), ql.Date(6,1,2017), ql.Date(6,2,2017),
                    ql.Date(6,3,2017), ql.Date(6,4,2017), ql.Date(6,5,2017), 
                    ql.Date(6,6,2017), ql.Date(6,7,2017), ql.Date(6,8,2017),
                    ql.Date(6,9,2017), ql.Date(6,10,2017), ql.Date(6,11,2017)]
strikes = [527.50, 560.46, 593.43, 626.40, 659.37, 692.34, 725.31, 758.28]
data = [
[0.37819, 0.34177, 0.30394, 0.27832, 0.26453, 0.25916, 0.25941, 0.26127],
[0.3445, 0.31769, 0.2933, 0.27614, 0.26575, 0.25729, 0.25228, 0.25202],
[0.37419, 0.35372, 0.33729, 0.32492, 0.31601, 0.30883, 0.30036, 0.29568],
[0.37498, 0.35847, 0.34475, 0.33399, 0.32715, 0.31943, 0.31098, 0.30506],
[0.35941, 0.34516, 0.33296, 0.32275, 0.31867, 0.30969, 0.30239, 0.29631],
[0.35521, 0.34242, 0.33154, 0.3219, 0.31948, 0.31096, 0.30424, 0.2984],
[0.35442, 0.34267, 0.33288, 0.32374, 0.32245, 0.31474, 0.30838, 0.30283],
[0.35384, 0.34286, 0.33386, 0.32507, 0.3246, 0.31745, 0.31135, 0.306],
[0.35338, 0.343, 0.33464, 0.32614, 0.3263, 0.31961, 0.31371, 0.30852],
[0.35301, 0.34312, 0.33526, 0.32698, 0.32766, 0.32132, 0.31558, 0.31052],
[0.35272, 0.34322, 0.33574, 0.32765, 0.32873, 0.32267, 0.31705, 0.31209],
[0.35246, 0.3433, 0.33617, 0.32822, 0.32965, 0.32383, 0.31831, 0.31344],
[0.35226, 0.34336, 0.33651, 0.32869, 0.3304, 0.32477, 0.31934, 0.31453],
[0.35207, 0.34342, 0.33681, 0.32911, 0.33106, 0.32561, 0.32025, 0.3155],
[0.35171, 0.34327, 0.33679, 0.32931, 0.3319, 0.32665, 0.32139, 0.31675],
[0.35128, 0.343, 0.33658, 0.32937, 0.33276, 0.32769, 0.32255, 0.31802],
[0.35086, 0.34274, 0.33637, 0.32943, 0.3336, 0.32872, 0.32368, 0.31927],
[0.35049, 0.34252, 0.33618, 0.32948, 0.33432, 0.32959, 0.32465, 0.32034],
[0.35016, 0.34231, 0.33602, 0.32953, 0.33498, 0.3304, 0.32554, 0.32132],
[0.34986, 0.34213, 0.33587, 0.32957, 0.33556, 0.3311, 0.32631, 0.32217],
[0.34959, 0.34196, 0.33573, 0.32961, 0.3361, 0.33176, 0.32704, 0.32296],
[0.34934, 0.34181, 0.33561, 0.32964, 0.33658, 0.33235, 0.32769, 0.32368],
[0.34912, 0.34167, 0.3355, 0.32967, 0.33701, 0.33288, 0.32827, 0.32432],
[0.34891, 0.34154, 0.33539, 0.3297, 0.33742, 0.33337, 0.32881, 0.32492]]

n=100

for i in range(n):
    s = time.time()
    implied_vols = ql.Matrix(len(strikes), len(expiration_dates))
    for i in range(implied_vols.rows()):
        for j in range(implied_vols.columns()):
            implied_vols[i][j] = data[j][i]
    black_var_surface = ql.BlackVarianceSurface(
        calculation_date, calendar, 
        expiration_dates, strikes, 
        implied_vols, day_count) 

    strike = 600.0
    expiry = 1.2 # years
    """
    black_var_surface.blackVol(expiry, strike)


    strikes_grid = np.arange(strikes[0], strikes[-1],10)
    expiry = 1.0 # years
    implied_vols = [black_var_surface.blackVol(expiry, s) 
                    for s in strikes_grid] # can interpolate here
    actual_data = data[11] # cherry picked the data for given expiry

    fig, ax = plt.subplots()
    ax.plot(strikes_grid, implied_vols, label="Black Surface")
    ax.plot(strikes, actual_data, "o", label="Actual")
    ax.set_xlabel("Strikes", size=12)
    ax.set_ylabel("Vols", size=12)
    legend = ax.legend(loc="upper right")
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    """

    plot_years = np.arange(0, 2, 0.1)
    plot_strikes = np.arange(535, 750, 1)

    X, Y = np.meshgrid(plot_strikes, plot_years)
    Z = np.array([black_var_surface.blackVol(float(y), float(x)) 
                for xr, yr in zip(X, Y) 
                    for x, y in zip(xr,yr) ]
                ).reshape(len(X), len(X[0]))
    total_seconds+=time.time() - s
    
    
print(f"total_seconds: {total_seconds}")
print(f"per sec: {n/total_seconds}")


#%%
surf = ax.plot_surface(X,Y,Z, rstride=1, cstride=1, cmap=cm.coolwarm, 
                linewidth=0.1)
fig.colorbar(surf, shrink=0.5, aspect=5)

#%%


implied_vols = [black_var_surface.blackVol(expiry, s) for s in strikes_grid]

target_date = calculation_date + ql.Period(1, ql.Years)
closest_date = min(expiration_dates, key=lambda d: abs(day_count.yearFraction(calculation_date, d) - expiry))
actual_strikes = strikes_data[closest_date]
actual_vols = vol_data[closest_date]


fig, ax = plt.subplots()
ax.plot(strikes_grid, implied_vols, label="Black Surface")
ax.plot(actual_strikes, actual_vols, "o", label="Actual")
ax.set_xlabel("Strikes", size=12)
ax.set_ylabel("Vols", size=12)
legend = ax.legend(loc="upper right")

plot_years = [day_count.yearFraction(calculation_date, d) for d in expiration_dates]
plot_strikes = all_strikes

#%%

X, Y = np.meshgrid(plot_strikes, plot_years)
Z = np.array([black_var_surface.blackVol(y, x)
              for xr, yr in zip(X, Y)
                  for x, y in zip(xr,yr)]
             ).reshape(len(X), len(X[0]))

print(Z)

#%%

fig = plt.figure()
ax = fig.add_subplot(projection='3d')



surf = ax.plot_surface(X,Y,Z, rstride=1, cstride=1, cmap=cm.coolwarm,
                linewidth=0.1)
fig.colorbar(surf, shrink=0.5, aspect=5)
plt.title("Volatility Surface with Max 6 Strikes/Expiry and Max 6 Expiries/Strike")
plt.show()

# Print some diagnostics
print("Strikes usage across expiries:")
for strike, count in strike_usage.items():
    print(f"Strike {strike}: appears in {count} expiries")

print("\nStrikes per expiry:")
for date in expiration_dates:
    print(f"{date}: {len(strikes_data[date])} strikes")
    
    
#%%%


v0 = 0.01; kappa = 0.2; theta = 0.02; rho = -0.75; sigma = 0.5

process = ql.HestonProcess(flat_ts, dividend_ts, 
                           ql.QuoteHandle(ql.SimpleQuote(spot)), 
                           v0, kappa, theta, sigma, rho)
model = ql.HestonModel(process)
engine = ql.AnalyticHestonEngine(model) 
heston_helpers = []
black_var_surface.setInterpolation("bicubic")
one_year_idx = 11 
date = expiration_dates[one_year_idx]
for j, s in enumerate(strikes):
    t = (date - calculation_date)
    p = ql.Period(t, ql.Days)
    #sigma = list(vol_data.values())[one_year_idx][j]
    print(t)
    sigma = black_var_surface.blackVol(t/365.25, s)
    print(sigma)
    helper = ql.HestonModelHelper(p, calendar, spot, s, 
                                  ql.QuoteHandle(ql.SimpleQuote(sigma)),
                                  flat_ts, 
                                  dividend_ts)
    helper.setPricingEngine(engine)
    heston_helpers.append(helper)

lm = ql.LevenbergMarquardt(1e-8, 1e-8, 1e-8)
model.calibrate(heston_helpers, lm, 
                 ql.EndCriteria(500, 50, 1.0e-8,1.0e-8, 1.0e-8))
theta, kappa, sigma, rho, v0 = model.params()


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any, Set
import math
from . import base
import numpy as np
from py_lets_be_rational.normaldistribution import norm_cdf, norm_pdf

import py_vollib_vectorized
from .base import OptionPricingModel

class BlackScholes(OptionPricingModel):
    def __init__(self, S: float, K: float, t: float, r: float, q: float, sigma: float, flag: int):
        self.S = S
        self.K = K
        self.t = t
        self.r = r
        self.q = q
        self.sigma = sigma
        self.flag = flag

    def _calculate_intermediates(self, greeks: Set[str]) -> Dict[str, float]:
        required = set()
        for g in greeks:
            if g in self._GREEK_REQUIREMENTS:
                required.update(self._GREEK_REQUIREMENTS[g])
        
        cache = {}
        if not required:
            return cache
        
        if "d1" in required:
            d1 = (math.log(self.S / self.K) + (self.r - self.q + 0.5 * self.sigma**2) * self.t)
            d1 /= (self.sigma * math.sqrt(self.t))
            cache["d1"] = d1
            
            if "phi_d1" in required:
                cache["phi_d1"] = norm_pdf(d1)
            if "N_flag_d1" in required:
                cache["N_flag_d1"] = norm_cdf(self.flag * d1)
        
        if "d2" in required and "d1" in cache:
            d2 = cache["d1"] - self.sigma * math.sqrt(self.t)
            cache["d2"] = d2
            if "N_flag_d2" in required:
                cache["N_flag_d2"] = norm_cdf(self.flag * d2)
        
        return cache

    def _compute_greeks(self, greeks: Set[str], cache: Dict[str, float]) -> Dict[str, float]:
        results = {}
        discount_q = math.exp(-self.q * self.t)
        discount_r = math.exp(-self.r * self.t)
        
        if "delta" in greeks:
            results["delta"] = self.flag * discount_q * cache["N_flag_d1"]
        if "gamma" in greeks:
            results["gamma"] = discount_q * cache["phi_d1"] / (self.S * self.sigma * math.sqrt(self.t))
        if "vega" in greeks:
            results["vega"] = self.S * discount_q * cache["phi_d1"] * math.sqrt(self.t)
        if "theta" in greeks:
            term1 = -self.S * discount_q * cache["phi_d1"] * self.sigma / (2 * math.sqrt(self.t))
            term2 = -self.flag * self.S * (self.r - self.q) * discount_q * cache["N_flag_d1"]
            term3 = self.flag * self.r * self.K * discount_r * cache["N_flag_d2"]
            results["theta"] = term1 + term2 + term3
        if "rho" in greeks:
            results["rho"] = self.flag * self.K * self.t * discount_r * cache["N_flag_d2"]
        if "price" in greeks:
            results["price"] = self.flag * (
                self.S * discount_q * cache["N_flag_d1"] - 
                self.K * discount_r * cache["N_flag_d2"]
            )
        return results
        
class Engine(base.ABCEuroEngine):
    def __init__(self, strike, expiry, flag, flag_int=None,
                 interest_rate_engine=None, dividend_rate_engine=None):
        super().__init__(strike, expiry, flag, flag_int, interest_rate_engine=interest_rate_engine, dividend_rate_engine=dividend_rate_engine)
        self.dividend_rate_engine = dividend_rate_engine
        self.IVOL_engine = self.IVOL
        self.greek_engine_all = self.get_all_greeks
        self.greeks_engines = {
                               "delta": self.delta,
                               "gamma": self.gamma,
                               "vega": self.vega,
                               "rho" : self.rho,
                               "theta" : self.theta
                              }
    @staticmethod
    def IVOL(price, S, K, t, r, flag, q=0, return_as='numpy', **kwargs):
        return py_vollib_vectorized.vectorized_implied_volatility(price, S, K, t, r, flag, q=q, model='black_scholes_merton',return_as=return_as)
    
    @staticmethod
    def delta(sigma, S, K, t, r, flag, q=0, return_as='numpy', **kwargs):
        d1 = (np.log(S/K) + (r - q + 0.5 * sigma**2) * t) / sigma / np.sqrt(t)
        return flag * np.exp(-q*t) * norm_cdf(-flag*d1, 0, 1)

    @staticmethod
    def gamma(sigma, S, K, t, r, flag, q=0, return_as='numpy', **kwargs):
        d1 = (np.log(S/K) + (r - q + 0.5 * sigma**2) * t) / (sigma * np.sqrt(t))
        return np.exp(-q * t) * norm_pdf(d1) / (S * sigma * np.sqrt(t))

    @staticmethod
    def vega(sigma, S, K, t, r, flag, q=0, return_as='numpy', **kwargs):
        d1 = (np.log(S/K) + (r - q + 0.5 * sigma**2) * t) / (sigma * np.sqrt(t))
        return S * np.exp(-q * t) * norm_pdf(d1) * np.sqrt(t)

    @staticmethod
    def theta(sigma, S, K, t, r, flag, q=0, return_as='numpy', **kwargs):
        d1 = (np.log(S/K) + (r - q + 0.5 * sigma**2) * t) / (sigma * np.sqrt(t))
        d2 = d1 - sigma * np.sqrt(t)
        term1 = - (S * np.exp(-q * t) * norm_pdf(d1) * sigma) / (2 * np.sqrt(t))
        term2 = flag * q * S * np.exp(-q * t) * norm_cdf(flag * d1)
        term3 = - flag * r * K * np.exp(-r * t) * norm_cdf(flag * d2)
        return term1 + term2 + term3

    @staticmethod
    def rho(sigma, S, K, t, r, flag, q=0, return_as='numpy', **kwargs):
        d1 = (np.log(S/K) + (r - q + 0.5 * sigma**2) * t) / (sigma * np.sqrt(t))
        d2 = d1 - sigma * np.sqrt(t)
        return flag * K * t * np.exp(-r * t) * norm_cdf(flag * d2)

    @staticmethod
    def get_all_greeks(sigma, S, K, t, r, flag, q=0, **kwargs):
        d1 = (np.log(S/K) + (r - q + 0.5 * sigma**2) * t) / (sigma * np.sqrt(t))
        d2 = d1 - sigma * np.sqrt(t)
        
        delta = flag * np.exp(-q * t) * norm_cdf(flag * d1)
        gamma = np.exp(-q * t) * norm_pdf(d1) / (S * sigma * np.sqrt(t))
        vega = S * np.exp(-q * t) * norm_pdf(d1) * np.sqrt(t)
        theta = - (S * np.exp(-q * t) * norm_pdf(d1) * sigma) / (2 * np.sqrt(t)) \
                + flag * q * S * np.exp(-q * t) * norm_cdf(flag * d1)\
                - flag * r * K * np.exp(-r * t) * norm_cdf(flag * d2)\
                
        rho = flag * K * t * np.exp(-r * t) * norm_cdf(flag * d2)
        return delta, gamma, vega, theta, rho
    

    @staticmethod
    def PC_parity(S, K, t, r, C=None, P=None, q=0, **kwargs):
        if P is not None:
            return P + S * np.exp(-q * t) - K * np.exp(-r * t)
        elif C is not None:
            return C - (S * np.exp(-q * t) - K * np.exp(-r * t))
        else:
            raise ValueError("Either C or P must be provided")
    
    @staticmethod
    def PC_parity_diff(S, K, t, r, C, P, q=0,):
        return C - P + S * np.exp(-q * t) - K * np.exp(-r * t)
    
    @staticmethod
    def get_implied_yield_from_PC(S, K, t, r, C=None, P=None, q=0, **kwargs):
        q = np.log((C - P + K*np.exp(-r*t))/S)/-t
        return q


from __future__ import annotations 
import numpy as np
import copy
from pyvol_terminal.plotting_engines import MetricEngine
from pyvol_terminal import metric_attributes
from classes import BaseDomain, Domain, SurfaceContainer, DataContainerManager, OptionChain, Points, surface, SmileContainer

def create_options_data_container_manager(df, config_instruments, config_interface):
    n_options = df.shape[0]
    name_index_map = {v: k for k, v in df["ticker"].to_dict().items()}
    base_domain = BaseDomain(df["strike"].values, df["expiry"].values, "Strike", "Expiry")    
    domain = Domain(base_domain)
    data_container_manager = DataContainerManager(MetricEngine(base_domain), config_instruments["price_types"])
    
    for px_type in config_instruments["price_types"]:
        domain = copy.deepcopy(domain)
        
        args = [n_options, name_index_map]
        for metric in metric_attributes.get_metric_attr():
            if metric == "OTM" or metric == "call_flag":
                val = False
            else:
                val = np.nan
            arr = np.full(n_options, val)
            args.append(arr) 
        raw = OptionChain(*args)
        
        if "Surface" in config_interface:
            scatter = Points(domain.x, domain.y, raw.ivol)
            surface = surface(base_domain,
                              scatter,
                              copy.deepcopy(config_interface["Surface"]["interpolation_config"]["engine"]),
                              config_interface["Surface"]["interpolation_config"]["n_x"],
                              config_interface["Surface"]["interpolation_config"]["n_y"])
        else:
            scatter=None
            surface=None
        if "Smile" in config_interface:
            smile = SmileContainer(base_domain,
                          raw,
                          copy.deepcopy(config_interface["Smile"]["interpolation_config"]["engine"]),
                          config_interface["Smile"]["interpolation_config"]["n_x"])
        else:
            smile=None
        data_container = SurfaceContainer(px_type,
                                       domain,
                                       raw,
                                       scatter,
                                       surface,
                                       smile)
        data_container.calculate_limits()
        data_container_manager.add_data_container(data_container)
    return data_container_manager

def create_matrices(df):
    x = df["strike"]
    y = df["expiry"]
    z = np.full(x.size, np.nan)
    instrument_names = df["ticker"].to_list()
    
    x_unique = np.unique(x)
    y_unique = np.unique(y)

    X_grid, Y_grid = np.meshgrid(x_unique, y_unique)
    Z_grid = np.full_like(X_grid, np.nan)
    name_index_map = {}
    for xi, yi, zi, name in zip(x, y, z, instrument_names):
        i = np.where(y_unique == yi)[0][0] 
        j = np.where(x_unique == xi)[0][0] 
        Z_grid[i, j] = zi
        name_index_map[name] = (i,j)

    return X_grid, Y_grid, Z_grid, name_index_map
    


def create_smile_data_containers(df, config_instruments, config_interface):
    n_options = df.shape[0]
    strike_grid, expiry_grid, price_grid, name_index_map = create_matrices(df)
    
    base_domain = BaseDomain(strike_grid, expiry_grid, "Strike", "Expiry")    
    domain = Domain(base_domain)
    data_container_manager = DataContainerManager(MetricEngine(base_domain), config_instruments["price_types"])
    
    for px_type in config_instruments["price_types"]:
        domain = copy.deepcopy(domain)
        
        args = [n_options, name_index_map]
        for metric in metric_attributes.get_metric_attr():
            if metric == "OTM" or metric == "call_flag":
                val = False
            else:
                val = np.nan
            arr = np.full(n_options, val)
            args.append(arr) 
        raw = OptionChain(*args)

    
        smile = SmileContainer(base_domain,
                    raw,
                    copy.deepcopy(config_interface["Smile"]["interpolation_config"]["engine"]),
                    config_interface["Smile"]["interpolation_config"]["n_x"])
        data_container = SurfaceContainer(px_type,
                                       domain,
                                       raw,
                                       smile)
        data_container.calculate_limits()
        data_container_manager.add_data_container(data_container)

    






def create_metric_arrays_for_view(n_options, price_types):
    metric_attr = metric_attributes.get_metric_attr()
    metric_pType_array_dict = {}
    idx_metric_array_dict = {n : {metric : None for metric in metric_attr} for n in range(n_options)}
    full_array = {}
    for metric, attr_dict in metric_attr.items():
        if metric == "OTM":
            val = False
        else:
            val = np.nan

        if attr_dict["multi_dim_flag"]:
            a1 = np.arange(1, n_options+1)
            a2 = 10*np.arange(1, n_options+1)
            a3 = 100 * np.arange(1, n_options+1)
            #array = np.column_stack((a1,a2,a3))
            array = np.full((n_options, len(price_types)), val)
            price_type_dict = {}
            
            for idx, px_type in enumerate(price_types):
                arr_col = array[:, idx]
                price_type_dict[px_type] = arr_col
                
            metric_pType_array_dict[metric] = price_type_dict
        else:
            #array = np.arange(1, n_options+1).reshape(n_options,1)
            array = np.full((n_options, 1), val)
            price_type_dict = {}
            for px_type in price_types:
                price_type_dict[px_type] = array
                
            metric_pType_array_dict[metric] = price_type_dict
        full_array[metric] = array
        for n in range(n_options):
            idx_metric_array_dict[n][metric] = array[n,:]
    return metric_pType_array_dict, idx_metric_array_dict, full_array


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from instruments.instruments import Option
 #   from 
    from instruments.utils import InstrumentManager

from PySide6 import QtWidgets, QtCore      
from abc import ABC, abstractmethod, ABCMeta
from ..gl_3D_graphing import meta
from pprint import pprint



class _ABCQWidgetMeta(type(QtCore.QObject), ABCMeta):...
class ABCQWidget(QtWidgets.QWidget, ABC, metaclass=_ABCQWidgetMeta):...

QSplitterMeta = type(QtWidgets.QSplitter)




class ABCInterface(QtWidgets.QMainWindow, metaclass=meta.QABCMeta):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.on_view=False

    
    @abstractmethod
    def get_calibration_slot(self) -> QtCore.Slot:...
    
    @QtCore.Slot()
    def update_view(self, calibrated_engines):
        if self.isVisible():
            self._internal_update_view(calibrated_engines)

    @abstractmethod
    def _internal_update_view(self):...


    
    ### RESUIRES_ABOUTTOCLOSE_TOO
        
    
#class ABCQWidget(QtWidgets.QWidget, ABC, metaclass=QObjectMeta):

QSplitterMeta = type(QtWidgets.QSplitter)



class ABCMainViewMeta222(_ABCQWidgetMeta, QSplitterMeta):
    
    def __init__(self, *args, **kwargs):
        self._intense_interaction=False
        self._interact_end_callbacks: List[Callable]=[]
        self._intenseInteractionContainer: Dict[str, bool]={}
        super().__init__(*args, **kwargs)
        
    @QtCore.Slot()
    def update_view(self, calibrated_engines):
        if self.isVisible():
            self._internal_update_view(calibrated_engines)

    def add_interact_end_callback(self, callback):
        self._interact_end_callbacks.append(callback)
        
    def remove_interact_end_callback(self, callback): 
        self._interact_end_callbacks.remove(callback)
 
    def processInteractionState(self, id, state: bool):
        self._intenseInteractionContainer[id] = state
        current_state = any(self._intenseInteractionContainer.values())
        if current_state != self._intense_interaction:
            self._intense_interaction = current_state
            for callback in self._interact_end_callbacks:
                callback(current_state)
                
    def addIntenseInteraction(self, id):
        self._intenseInteractionContainer[id] = False
    
    def removeIntenseInteraction(self, id):
        del self._intenseInteractionContainer[id]

    @abstractmethod
    def _internal_update_view(self):...


class ABCMainViewQSplitter(QtWidgets.QSplitter, metaclass=meta.QABCMeta):
    
    def __init__(self, *args, **kwargs):
        self._intense_interaction=False
        self._interact_end_callbacks: List[Callable]=[]
        self._intenseInteractionContainer: Dict[str, bool]={}
        super().__init__(*args, **kwargs)
        
    @QtCore.Slot()
    def update_view(self, calibrated_engines):
        if self.isVisible():
            self._internal_update_view(calibrated_engines)

    def add_interact_end_callback(self, callback):
        self._interact_end_callbacks.append(callback)
        
    def remove_interact_end_callback(self, callback): 
        self._interact_end_callbacks.remove(callback)
 
    def processInteractionState(self, id, state: bool):
        self._intenseInteractionContainer[id] = state
        current_state = any(self._intenseInteractionContainer.values())
        if current_state != self._intense_interaction:
            self._intense_interaction = current_state
            for callback in self._interact_end_callbacks:
                callback(current_state)
                
    def addIntenseInteraction(self, id):
        self._intenseInteractionContainer[id] = False
    
    def removeIntenseInteraction(self, id):
        del self._intenseInteractionContainer[id]
        
    @abstractmethod
    def _internal_update_view(self):...


class ABCMainViewQTableWidget(QtWidgets.QTableWidget, metaclass=meta.QABCMeta):
    
    def __init__(self, *args,**kwargs):
        self._intense_interaction=False
        self._interact_end_callbacks: List[Callable]=[]
        self._intenseInteractionContainer: Dict[str, bool]={}
        super().__init__(*args, **kwargs)
        
    @QtCore.Slot()
    def update_view(self, calibrated_engines):
        if self.isVisible():
            self._internal_update_view(calibrated_engines)

    def add_interact_end_callback(self, callback):
        self._interact_end_callbacks.append(callback)
        
    def remove_interact_end_callback(self, callback): 
        self._interact_end_callbacks.remove(callback)
 
    def processInteractionState(self, id, state: bool):
        self._intenseInteractionContainer[id] = state
        current_state = any(self._intenseInteractionContainer.values())
        if current_state != self._intense_interaction:
            self._intense_interaction = current_state
            for callback in self._interact_end_callbacks:
                callback(current_state)
                
    def addIntenseInteraction(self, id):
        self._intenseInteractionContainer[id] = False
    
    def removeIntenseInteraction(self, id):
        del self._intenseInteractionContainer[id]


    @abstractmethod
    def _internal_update_view(self):...



class ABCSettings(QtWidgets.QWidget, metaclass=meta.QABCMeta):
    def __init__(self, *args,**kwargs):
        self._parametersUpdatedCallbacks=[]
        super().__init__(*args, **kwargs)

    def addParametersUpdatedCallback(self, callback):
        self._parametersUpdatedCallbacks.append(callback)
    
    def removeParametersUpdatedCallback(self, callback):
        self._parametersUpdatedCallbacks.remove(callback)
    
    def parametersUpdated(self, new, previous=None):
        for callback in self._parametersUpdatedCallbacks:
            callback(new, previous)


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from instruments.instruments import Option, Spot
    from ..graphics_items.GL3DGraphicsItems import GlPlotItemMixin
    from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem

from PySide6 import QtCore, QtGui
import numpy as np
import numpy as np
from dataclasses import dataclass, InitVar, field
from PySide6 import QtCore
from ..graphics_items import GL3DAxisItem
from .. import utils
from pprint import pprint   
from pyqtgraph import opengl

@dataclass(slots=True)
class PlotView:
    axis: str
    axis_idx: int
    min: float = np.nan
    max: float = np.nan
    min_UB: float = np.nan
    max_LB: float = np.nan
    
    
    def normalise(self, values):
        return (values - self.min) / (self.max - self.min)
    
    def unnormalise(self, values):
        return values * (self.max - self.min) + self.min
    
    def update_min(self, min):
        self.min = min
        self._get_min_UB()
    
    def update_max(self, max):
        self.max = max
        self._get_max_LB()
    
    def _get_min_UB(self):
        if not np.isnan(self.max):
            d = self.max - self.min
            self.min_UB = self.min + 0.4 * d
            
    def _get_max_LB(self):
        if not np.isnan(self.min):
            d = self.max - self.min
            self.max_LB = self.max - 0.4 * d

    def is_new_limit(self, limit_type, value):
        if np.isnan(getattr(self, limit_type)):
            return True
        else:
            if limit_type == "min":
                if np.isnan(self.min_UB) and not np.isnan(self.max):
                    self._get_min_UB()
                if value < self.min or value > self.min_UB:
                    return True
                else:
                    return False
            else:
                if np.isnan(self.max_LB) and not np.isnan(self.min):
                    self._get_max_LB()
                if value > self.max or value < self.max_LB:
                    return True
                else:
                    return False
    
    def mask_from_data(self, data):
        return (data >= self.min) & (data <= self.max)

    
    
    
    
class CustomGLGraphicsItem(opengl.GLGraphicsItem.GLGraphicsItem):
    def __init__(self, *args, **kwargs):
        self.__children=[]
        super().__init__(*args, **kwargs)

    def addChildItem(self, item):
        self.__children.append(item)
        
    def childItems(self):
        return list(self.__children)
    


class ViewBox(CustomGLGraphicsItem):
    sigXRangeChanged = QtCore.Signal(float, float)
    sigYRangeChanged = QtCore.Signal(float, float)
    sigZRangeChanged = QtCore.Signal(float, float)

    
    def __init__(self,
                 norm_dims=[1,1,1],
                 padding=0.1,
                 *args, **kwargs):
        self.norm_dims=norm_dims
        self.padding=padding*np.array(norm_dims)
        self.min_vals, self.max_vals = None, None
        self.vb_min, self.vb_max = None, None
        self.viewRange = [[0, 1] for _ in range(3)]
        self.transform_matrix=None
        super().__init__(*args, **kwargs)
    
    def addChildItem(self, item):
        super().addChildItem(item)
        if hasattr(item, "linktoView"):
            item.linktoView(self)
        self.normalize_scene()
        item.setTransform(self.transform_matrix)
    
    def transformChildItems(self):
        for child in self.childItems():
            child.setTransform(self.transform_matrix)


    def normalize_scene(self):
        min_vals, max_vals = utils.compute_scene_bbox(self.childItems())
        if self.min_vals is None or any(min_vals < self.vb_min) or any(max_vals > self.vb_max):
            self.min_vals = min_vals
            self.max_vals = max_vals
            self.vb_min = min_vals - self.padding
            self.vb_max = max_vals + self.padding
            
            self.viewRange[0] = [self.vb_min[0], self.vb_max[0]]
            self.viewRange[1] = [self.vb_min[1], self.vb_max[1]]
            self.viewRange[2] = [self.vb_min[2], self.vb_max[2]]

            self.createTransformMatrix()
            self.transformChildItems()

    def setRange(self, xRange=None, yRange=None, zRange=None, padding=None):
        if all((xRange is None,
                yRange is None,
                zRange is None)):
            return 
        x,y,z=[False]*3
        
        if not xRange is None:
            x=True
            self.vb_min[0], self.vb_max[0] = xRange[0], xRange[1]
            self.viewRange[0]=xRange
            self.sigXRangeChanged.emit(*xRange)
            
                

        if not yRange is None:
            y=True
            self.vb_min[1], self.vb_max[1] = yRange[0], yRange[1]
            self.viewRange[1]=yRange
            self.sigYRangeChanged.emit(*yRange)
            
        
        if not zRange is None:
            z=True
            self.vb_min[2], self.vb_max[2] = zRange[0], zRange[1]
            self.viewRange[2]=zRange
            self.sigZRangeChanged.emit(*zRange)
            
        self.createTransformMatrix()
        
        for child in self.childItems():
            if hasattr(child, "filter_data"):
                child.filter_data(self)
            else:
                print(child)
                print(child.childItems())
            child.setTransform(self.transform_matrix)

    def createTransformMatrix(self):
        ranges = self.vb_max - self.vb_min
        ranges[ranges == 0] = 1.0
        
        self.transform_matrix = QtGui.QMatrix4x4()

        self.transform_matrix.scale(1/ranges[0], 1/ranges[1], 1/ranges[2])
        self.transform_matrix.translate(-self.vb_min[0], -self.vb_min[1], -self.vb_min[2])



from __future__ import annotations
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from pyvol_terminal.interfaces.volatility_surface.pyvol_GL3DViewWidget import SurfaceViewWidget
    from .GL3DAxisItem import GL3DViewBox
    from ..widgets.GL3DViewWidget import GL3DViewWidget
    

from pyqtgraph import opengl, Vector
from PySide6 import QtCore, QtGui, QtWidgets
import numpy as np
import math
from ..meta import QABCMeta, abc
from pyqtgraph.opengl import GLGraphicsItem
from OpenGL import GL  
from pprint import pprint   
from .GL3DGraphicsItems import MixedGL3DTextItem
from pyqtgraph import Transform3D

class GL3DTextItem(MixedGL3DTextItem):
    eye_mat = np.eye(4)
    def __init__(self,
                 anchor=(0, 0),
                 resizeToWorld=False,
                 resizeTolerance=0,
                 parentItem=None, 
                 **kwargs
                 ):    
        
        self.lastRect=None
        self.posF=None
        self.resizeTolerance=resizeTolerance
        glopts = kwargs.pop('glOptions', 'additive',)
        self._initDistance, self._initPointSize, self._lastSetDistance=None, None, None
        self._resizeToWorld=resizeToWorld
        self._anchor=anchor
        self.width=None
        self.p=None
        
        super().__init__(parentItem=parentItem, **kwargs)
        self.setGLOptions(glopts)

        
    def compute_projection(self):
        rect = QtCore.QRectF(self.view().rect())
        ndc_to_viewport = QtGui.QMatrix4x4()
        ndc_to_viewport.viewport(rect.left(), rect.bottom(), rect.width(), -rect.height())
        return ndc_to_viewport * self.mvpMatrix()
    
    def linkWidget(self, parent: SurfaceViewWidget):
        super().setParent(parent)
        
    def itemChange(self, change, value):
        ret = super().itemChange(change, value)
        if change == self.GLGraphicsItemChange.ItemViewHasChanged:
            self._initDistance, self._initPointSize=None, None
        return ret

    def _updateSize(self):
        if self._initDistance is None:
            view = self.view()
            if not view is None and hasattr(view, "cameraPosition"):
                distance = GL3DTextItem.distance_between_point_perpendicular_plane(self.pos, self.view().centerPositionNumpy(), self.view().cameraPositionNumpy())
                self._initDistance = distance
                self._initPointSize = self.font.pointSize()
                self._lastSetDistance = distance
                
        else:
                current_distance = GL3DTextItem.distance_between_point_perpendicular_plane(self.pos, self.view().centerPositionNumpy(), self.view().cameraPositionNumpy())
            #if abs(current_distance / self._lastSetDistance - 1) > self.resizeTolerance:
                self._lastSetDistance = current_distance
                r = self._initDistance / current_distance
                newPointSize = r * self._initPointSize
                self.font.setPointSizeF(newPointSize)
            
    def setData(self, **kwargs):
        if "resizeToWorld" in kwargs:
            self.setResizeToWorld(kwargs.pop("resizeToWorld"))
        if "anchor" in kwargs:
            self.setAnchor(kwargs.pop("anchor"))
        super().setData(**kwargs)

    @classmethod
    def distance_between_point_perpendicular_plane(cls, point, p2, p3):
        diff_OC = p2 - p3
        if np.linalg.norm(diff_OC) == 0:
            raise ValueError("Camera is already at the origin. No unique line of sight.")
        
        u_hat = diff_OC / np.linalg.norm(diff_OC)
        return np.dot(point - p3, u_hat)

    def _updatePositionFromAnchor(self, pos: QtGui.QVector3D):
        x, y = pos.x(), pos.y()
        font_metrics = QtGui.QFontMetrics(self.font)
        self.lastRect = font_metrics.tightBoundingRect(self.text)

        height = self.lastRect.height()
        self.width = self.lastRect.width()
        x = x - self._anchor[0] * self.width if self._anchor[0] == 1 else x
        y = y + self._anchor[1] * height if self._anchor[1] == 1 else y
        pos.setX(x)
        pos.setY(y)
        
        
    def setAnchor(self, anchor: Tuple[float, float]) -> None:
        self._anchor = anchor
        
    def setResizeToWorld(self, flag):
        self._resizeToWorld=flag    
        
    def anchor(self): return self._anchor
    def resizeToWorld(self): return self._resizeToWorld


    def _paintHelper(self):...

    def paint(self):
        if len(self.text) < 1:
            return
        self.setupGLState()
        
        if self.resizeToWorld():
            self._updateSize()
            
        project = self.compute_projection()
        vec3 = QtGui.QVector3D(*self.pos)
        self.posF = project.map(vec3).toPointF()        
        
        self._updatePositionFromAnchor(self.posF)
        
        painter = QtGui.QPainter(self.view())
        painter.setPen(self.color)
        painter.setFont(self.font)
        painter.setRenderHints(QtGui.QPainter.RenderHint.Antialiasing | QtGui.QPainter.RenderHint.TextAntialiasing)
        painter.drawText(self.posF, self.text)
        painter.end()




from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from instruments.instruments import Option
    from instruments.utils import InstrumentManager
    from pandas import DataFrame
    from data_classes.classes import BaseDomain, Domain, OptionChain, VolatilityData, VolVector
    from quantities.engines import MetricAxisEngine
    
    from ..engines.interpolation_engines import Abstract3DInterpolator

import numpy as np
import copy
from pyvol_terminal import metric_attributes
from . import classes#, containers
from ..quantities import engines


def _create_scatter_surface_dataclasses(px_type: str,
                                        domain: Domain,
                                        n_options,
                                        name_index_map,
                                        metric_engine: MetricAxisEngine,
                                        n_ticks: Tuple[int, int]
                                        ) -> VolatilityData:
    raw_opt_kwargs = {"n_instruments" : n_options,
                      "name_index_map" : name_index_map,
                      }
    
    greeks = metric_attributes.greeks_attr()
    moneyness=metric_attributes.moneyness_attr()
    greeks_container={}
    moneyness_container={}
    for metric in metric_attributes.get_metric_attr_all():
        if metric == "OTM" or metric == "call_flag":
            val = False
            dtype=bool
        else:
            val=np.nan
            dtype=float
        arr = np.full(n_options, val, dtype=dtype)
        if metric in greeks:
            greeks_container[metric]=arr
        elif metric in moneyness:
            moneyness_container[metric]=arr
        else:
            raw_opt_kwargs[metric]=arr
    
    greeks_container = classes.Greeks(**greeks_container)
    moneyness_container = classes.Moneyness(**moneyness_container)
    
    raw_opt_kwargs["greeks"] = greeks_container
    raw_opt_kwargs["moneyess"] = moneyness_container
    del raw_opt_kwargs["px_type"]
    option_chain = classes.OptionChain(**raw_opt_kwargs)
    
    vol_vector = classes.VolVector(option_chain=option_chain,
                                   strikes=domain.x,
                                   expiry=domain.y,
                                   metric_engine=metric_engine,
                                   )
    points_kwargs = {"x" : domain.x,
                     "y" : domain.y,
                     "z" : option_chain.ivol,
                     }
    points = classes.Points(**points_kwargs)

    surface_kwargs = {"vol_vector" : vol_vector,
                      "n_x" : n_ticks[0],
                      "n_y" : n_ticks[1],
                      }
    surface = classes.Surface(**surface_kwargs)                   
    skew = classes.Slice(vol_vector=vol_vector,
                    #    parent_dataclass=surface,
                         axis_3D_directions="xz",
                         )
    term = classes.Slice(vol_vector=vol_vector,
                     #    parent_dataclass=surface,
                        axis_3D_directions="yz",
                        )
    volatility_data = classes.VolatilityData(px_type=px_type,
                                            vol_vector=vol_vector,
                                            points=points,
                                            surface=surface,
                                            skew=skew,
                                            term=term)
    return volatility_data
    #return raw, scatter, surface, skew, term


def create_volatility_data_from_vol_vect(vol_vect_container: Dict[str, VolVector],
                                         n_ticks: Tuple[int, int]=(30,30),
                                         ) -> Dict[str, VolatilityData]:
    volatility_data_container = {}
    for px_type, vol_vect in vol_vect_container.items():
        domain = vol_vect.domain
        x, y, z = vol_vect.data()
        points_kwargs = {"x" : x,
                        "y" : y,
                        "z" : z,
                        "domain_vect" : domain,
                 #       "strikes" : vol_vect.strikes,
                 #       "expiry" : vol_vect.expiry,
                        }
        points = classes.Points(**points_kwargs)
    
        surface_kwargs = {"domain_vect" : domain,
                          "n_x" : n_ticks[0],
                          "n_y" : n_ticks[1],
                          }
        
        
        surface = classes.Surface(**surface_kwargs)
                                
        skew = classes.Slice(domain_vect=domain,
                          #   parent_dataclass=surface,
                             axis_3D_directions="xz",
                             name="skew"
                             )
        
        term = classes.Slice(domain_vect=domain,
                           #  parent_dataclass=surface,
                             axis_3D_directions="yz",
                             name="term"
                             )
        surface.add_update_callback(skew.update_from_surface)
        surface.add_update_callback(term.update_from_surface)

        volatility_data = classes.VolatilityData(px_type=px_type,
                                                vol_vector=vol_vect,
                                                scatter=points,
                                                surface=surface,
                                                skew=skew,
                                                term=term)
        volatility_data_container[px_type]=volatility_data
    return volatility_data_container

def create_parent_option_container(all_price_types: List[str],
                                   df: DataFrame,
                                   instrument_manager: InstrumentManager,
                                   ) -> Dict[str, OptionChain]:
    n_options = df.shape[0]
    
    name_index_map = {v: k for k, v in df["ticker"].to_dict().items()}
    raw_opt_kwargs = {"n_instruments" : n_options,
                      "name_index_map" : name_index_map,
                      }
    option_chain_container={}
    for px_type in all_price_types:
        greeks = metric_attributes.greeks_attr()
        moneyness=metric_attributes.moneyness_attr()
        greeks_container={}
        moneyness_container={}
        for metric in metric_attributes.get_metric_attr_all():
            if metric == "OTM" or metric == "call_flag":
                value = False
                dtype=bool
            else:
                value=np.nan
                dtype=float
            if metric not in ["strike", "expiry"]:
                arr = np.full(n_options, value, dtype=dtype)
            else:
                arr = df[metric].values
            if metric in greeks:
                greeks_container[metric]=arr
            elif metric in moneyness:
                moneyness_container[metric]=arr
            else:
                raw_opt_kwargs[metric]=arr
        
        greeks_container = classes.Greeks(**greeks_container)
        moneyness_container = classes.Moneyness(**moneyness_container)
        
        raw_opt_kwargs["greeks"] = greeks_container
        raw_opt_kwargs["moneyess"] = moneyness_container
        raw_opt_kwargs["px_type"]=px_type
        """
        option_chain = classes.OptionChain(**raw_opt_kwargs)

        for option_object in instrument_manager.options_instrument_container.objects.values():
            option_object.add_array_callback(option_chain.update_from_object)
        
        option_chain_container[px_type]=option_chain
        """
        for px_type in all_price_types:
            option_chain_container[px_type]={px_type : None}
        
    return option_chain_container

def create_option_chain(option_ptype: str,
                        underlying_ptype: str
                        ) -> OptionChain:
        greeks = metric_attributes.greeks_attr()
        moneyness=metric_attributes.moneyness_attr()
        greeks_container={}
        moneyness_container={}
        for metric in metric_attributes.get_metric_attr_all():
            if metric == "OTM" or metric == "call_flag":
                value = False
                dtype=bool
            else:
                value=np.nan
                dtype=float
            if metric not in ["strike", "expiry"]:
                arr = np.full(n_options, value, dtype=dtype)
            else:
                arr = df[metric].values
            if metric in greeks:
                greeks_container[metric]=arr
            elif metric in moneyness:
                moneyness_container[metric]=arr
            else:
                raw_opt_kwargs[metric]=arr
        
        greeks_container = classes.Greeks(**greeks_container)
        moneyness_container = classes.Moneyness(**moneyness_container)
        
        raw_opt_kwargs["greeks"] = greeks_container
        raw_opt_kwargs["moneyess"] = moneyness_container
        raw_opt_kwargs["px_type"]=px_type
        """
        option_chain = classes.OptionChain(**raw_opt_kwargs)

        for option_object in instrument_manager.options_instrument_container.objects.values():
            option_object.add_stucture_callback(option_chain.update_from_object)
        
        option_chain_container[px_type]=option_chain
        """
        for px_type in all_price_types:
            option_chain_container[px_type]={px_type : None}



def create_vol_vect_container(option_chain_container: Dict[str, Dict[str, OptionChain]],
                              ) -> Dict[str, VolVector]:
    option_chain = option_chain_container[list(option_chain_container.keys())[0]]
    strike = option_chain.strike
    expiry = option_chain.expiry
    base_domain = classes.BaseDomain(strike, expiry, "Strike", "Expiry")    
    
    
    domain = classes.Domain(base_domain)
    vol_vector_container={}
    
    for px_type, inner_opt_chain_container in option_chain_container.items():
        for px_type , option_chain in inner_opt_chain_container.items():
    
            metric_engine = engines.MetricAxisEngine(base_domain.strike,
                                                    base_domain.expiry,
                                                    "Strike",
                                                    "Expiry",
                                                    "Implied Volatility"
                                                    )
            domain_vect = classes.DomainVect(base_domain=base_domain)
            vol_vector = classes.VolVector(option_chain=option_chain,
                                           strikes=domain.x,
                                           expiry=domain.y,
                                           domain=domain_vect,
                                           metric_engine=metric_engine,
                                           )
            vol_vector_container[px_type]={px_type : vol_vector}
    return vol_vector_container
        


def create_vol_vect_container2222222(option_chain_container: Dict[str, OptionChain],
                              ) -> Dict[str, VolVector]:
    option_chain = option_chain_container[list(option_chain_container.keys())[0]]
    strike = option_chain.strike
    expiry = option_chain.expiry
    base_domain = classes.BaseDomain(strike, expiry, "Strike", "Expiry")    
    
    
    domain = classes.Domain(base_domain)
    
    metric_engine = engines.MetricAxisEngine(base_domain.strike,
                                             base_domain.expiry,
                                             "Strike",
                                             "Expiry",
                                             "Implied Volatility"
                                             )
    vol_vector_container={}
    for px_type, option_chain in option_chain_container.items():
        domain_vect = classes.DomainVect(base_domain=base_domain)
        vol_vector = classes.VolVector(option_chain=option_chain,
                                       strikes=domain.x,
                                       expiry=domain.y,
                                       domain=domain_vect,
                                       metric_engine=metric_engine,
                                       )
        vol_vector_container[px_type]=vol_vector
    return vol_vector_container



def create_surface_dataclasses(all_price_types: List[str],
                               df: DataFrame,
                               domain: Domain,
                               option_objects: List[Option],
                               metric_engine,
                               n_ticks: Tuple[int, int]
                               ) -> Dict[str, VolatilityData]:
    n_options = df.shape[0]
    name_index_map = {v: k for k, v in df["ticker"].to_dict().items()}
    
    dataclass_type_container = {"raw" : None,
                                "scatter" : None,
                                "surface" : None,
                                "skew" : None,
                                "term" : None
                                }
    
    volatility_data_container = {px_type : copy.deepcopy(dataclass_type_container) for px_type in all_price_types}

    for px_type in all_price_types:
        domain = copy.deepcopy(domain)
        volatility_data = _create_scatter_surface_dataclasses(px_type,
                                                                domain,
                                                                n_options,
                                                                name_index_map,
                                                                metric_engine,
                                                                n_ticks)
        volatility_data_container[px_type]=volatility_data

        for option_object in option_objects:
            option_object.add_child_derivative_callback(volatility_data.option_chain.update_from_instrument_object)

    return volatility_data_container
    
def create_domains(df):    
    base_domain = classes.BaseDomain(df["strike"].values, df["expiry"].values, "Strike", "Expiry")    
    domain = classes.Domain(base_domain)
    return base_domain, domain
    
def create_slice_container(df, price_types, config):
    n_options = df.shape[0]
    name_index_map = {v: k for k, v in df["ticker"].to_dict().items()}
    
    KT_idx_map={}
    for idx, kt in enumerate(zip(df["strike"].values, df["expiry"].values)):
        KT_idx_map[kt] = idx

    base_domain = classes.BaseDomain(df["strike"].values, df["expiry"].values, "Strike", "Expiry")    
    domain = classes.Domain(base_domain)
    metric_axis_engine=engines.MetricAxisEngine(base_domain.strike,
                                                   base_domain.expiry,
                                                   "Strike",
                                                   "Expiry",
                                                   "Implied Volatility"
                                                   )
    interpolator_engine = config["interpolation_engine_container"]
    
    raw_objects={}
    slice_objects={}
    dataclass_objects = {}
    domain_objects = {}

    displayed_price_types = ["mid"]
    for px_type in price_types:
        domain, raw, scatter, base_dataclass_kwargs, surface_kwargs = _create_surface_dataclass_args(px_type,
                                                                                        domain,
                                                                                        config["interpolation_engine_container"],
                                                                                        n_options,
                                                                                        name_index_map)
        surface = classes.Surface(**surface_kwargs)
        slice_kwargs = copy.deepcopy(base_dataclass_kwargs)
        slice_kwargs["surface"] = surface
        slice_kwargs["base_domain"] = base_domain
        domain_objects[px_type] = domain.x_vect if config["axis_direction"] == "xz" else domain.y_vect
        slice_kwargs["surface_axis_directions"] = config["axis_direction"]
        slice_kwargs["scatter"] = scatter
        
        
        slices = classes.Slice(**slice_kwargs)
        slice_objects[px_type] = slices
        raw_objects[px_type] = raw
        dataclass_objects[px_type] = {"slices" : slices}

    args = [price_types,
            displayed_price_types,
            df["ticker"].to_list(),
            metric_axis_engine,
            domain,
            raw_objects,
            [slice_objects],
            interpolator_engine,
            config["axis_direction"]]
    
    kwargs = {"all_price_types" : price_types,
              "displayed_price_types" : displayed_price_types,
              "instrument_names" : df["ticker"].to_list(),
              "metric_axis_engine" : metric_axis_engine,
              "domain" : domain,
              "raw" : raw_objects,
              "dataclasses" : dataclass_objects,
              "interpolator_engine" : interpolator_engine,
              "axis_direction" : config["axis_direction"],
              "slice_container" : slice_objects,
              "domain_vec" : domain_objects}
    
    
    
    return containers.SliceContainer(**kwargs)


def create_smile_container(df, price_types, config):
    return create_slice_container(df, price_types, config, 0)

def create_term_container(df, price_types, config):
    return create_slice_container(df, price_types, config, 1)



def _create_surface_dataclass_args(px_type, instrument_manager, domain, interpolation_engine, n_options, name_index_map):
        domain = copy.deepcopy(domain)
        args = [n_options, name_index_map]
        for metric in metric_attributes.get_metric_attr():
            if metric == "OTM" or metric == "call_flag":
                val = False
                dtype = bool
            else:
                val = np.nan
                dtype=float
            arr = np.full(n_options, val, dtype=dtype)
            args.append(arr) 
        raw = classes.OptionChain(*args)
        
        for option_object in instrument_manager.options_instrument_container.get_objects():
            option_object.add_update_callback(lambda px_type: option_object.get_all_metrics_price_type(px_type))
        
        base_dataclass_kwargs = {"px_type" : px_type,
                                 "x" : domain.x,
                                 "y" : domain.y,
                                 "z" : raw.ivol,
                                 }
        scatter = classes.Points(**base_dataclass_kwargs)
        
        surface_kwargs = copy.deepcopy(base_dataclass_kwargs)
        
        surface_kwargs["interpolator"] = copy.deepcopy(interpolation_engine["engine"])

        args=[px_type,domain.x, domain.y, raw.ivol, copy.deepcopy(interpolation_engine["engine"])]
        return domain, raw, scatter, base_dataclass_kwargs, surface_kwargs



def create_surface_container(df, instrument_manager, price_types, config, include_scatter=False):
    n_options = df.shape[0]
    name_index_map = {v: k for k, v in df["ticker"].to_dict().items()}
    
    base_domain = classes.BaseDomain(df["strike"].values, df["expiry"].values, "Strike", "Expiry")    
    domain = classes.Domain(base_domain)
    
    raw_objects = {}
    domain_objects = {}
    surface_objects = {}
    scatter_objects = {}
    for px_type in price_types:
        domain, raw, scatter, _, surface_kwargs = _create_surface_dataclass_args(px_type,
                                                                                 instrument_manager,
                                                                    domain,
                                                                    config["interpolation_engine_container"],
                                                                    n_options,
                                                                    name_index_map)
        surface = classes.Surface(**surface_kwargs)
        raw_objects[px_type] = raw
        domain_objects[px_type] = domain
        scatter_objects[px_type] = scatter
        surface_objects[px_type] = surface
    
    displayed_price_types = ["mid"]
    metric_axis_engine=engines.MetricAxisEngine(base_domain.strike, base_domain.expiry, "Strike", "Expiry", "Implied Volatility")
    
    kwargs = {"all_price_types" : price_types,
              "displayed_price_types" : displayed_price_types,
              "instrument_names" : df["ticker"].to_list(),
              "metric_axis_engine" : metric_axis_engine,
              "domain" : domain,
              "raw" : raw_objects,
              "surface_dataclasses" : surface_objects,
              "scatter_dataclasses" : scatter_objects}    
    return containers.SurfaceContainer(**kwargs)

def create_matrices(df):
    x = df["strike"]
    y = df["expiry"]
    z = np.full(x.size, np.nan)
    instrument_names = df["ticker"].to_list()
    
    x_unique = np.unique(x)
    y_unique = np.unique(y)

    X_grid, Y_grid = np.meshgrid(x_unique, y_unique)
    Z_grid = np.full_like(X_grid, np.nan)
    name_index_map = {}
    for xi, yi, zi, name in zip(x, y, z, instrument_names):
        i = np.where(y_unique == yi)[0][0] 
        j = np.where(x_unique == xi)[0][0] 
        Z_grid[i, j] = zi
        name_index_map[name] = (i,j)

    return X_grid, Y_grid, Z_grid, name_index_map


def create_data_containers(df, instrument_manager, price_types, config_interface):
    data_container_manager = {}
    for interface, config in config_interface.items():
        match interface:
            case "Vol Table":
                container=create_surface_container(df, price_types, config)
            case "OMON":
                container=create_surface_container(df, price_types, config)
            case "Surface":
                container=create_surface_container(df, instrument_manager, price_types, config, include_scatter=True)
            case "Smile":
                container=create_slice_container(df, price_types, config)
            case "Term":
                container=create_slice_container(df, price_types, config)
        data_container_manager[interface] = container
    return data_container_manager




from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from .GL3DViewBox import GL3DViewBox 
    
from pyqtgraph import opengl
from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
from . import AbstractGLPlotItem
from pyqtgraph.Qt import QT_LIB
import importlib
from .GLGraphicsItem import AbstractGLGraphicsItem
from .AbstractGLPlotItem import AbstractGLPlotItem  


if QT_LIB in ["PyQt5", "PySide2"]:
    QtOpenGL = QtGui
else:
    QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")

__all__ = ['GLScatterPlotItem']


opengl.GLScatterPlotItem.__bases__ = (AbstractGLPlotItem,)

class GLScatterPlotItem(opengl.GLScatterPlotItem, AbstractGLPlotItem):        
    sigPlotChanged = QtCore.Signal(object)       
    def __init__(self, parentItem=None, **kwds):
        print(f"inti: {kwds}")
        AbstractGLPlotItem.__init__(self, **kwds)
        glopts = kwds.pop('glOptions', 'additive')
        self.setGLOptions(glopts)
        self.pos = None
        self.size = 10
        self.color = [1.0,1.0,1.0,0.5]
        self.pxMode = True

        self.m_vbo_position = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_vbo_color = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_vbo_size = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.vbos_uploaded = False
        self.setData = AbstractGLPlotItem.setData.__get__(self, type(self))  
        self.setParentItem(parentItem)
        self.setData(**kwds)

    def _internal_update(self):
        return opengl.GLScatterPlotItem.update()
    
        
    def dataBounds(self,
                   ax: int,
                   frac=(1,1),
                   orthoRange=None
                   ) -> Tuple[float, float, float]:

        if not self.visible() or self._dataset is None:
            return None, None
        else:
            
            if frac >= 1.0 and orthoRange is None and not self._boundsCache[ax] is None:
                return self._boundsCache[ax]
            data = self._dataset.data()
            if data is None:
                return None, None
            if orthoRange is None:
                mask = ~np.isnan(data[ax])
            else:
                mask = (data[ax] >= orthoRange[0]) & (data[ax] <= orthoRange[1])
            data = data[ax][mask]
            self._boundsCache[ax] = [np.nanmin(data), np.nanmax(data)]
            
            return self._boundsCache[ax]
    
    
    def _setDataHelper(self, pos=None, color=None):
        print(f"\n_setDataInternal\n")
        kwargs={}
        if pos is None:
            #self._dataset=None
            if color is None:
                opengl.GLScatterPlotItem.setData(self, color=color)
                return 
        else:
            self._dataset = self.plotdataset_cls(pos[:,0], pos[:,1], pos[:,2])
            #self._dataset.update(pos)
            #self._dataset = PlotDataset(pos[:,0], pos[:,1], pos[:,2])
            self._datasetDisplay = self._getDisplayDataset()
            kwargs["pos"] = self._datasetDisplay.data()
        if not color is None:
            kwargs["color"]=color
        

        #self.blockUpdate(True)
        print(kwargs)
        opengl.GLScatterPlotItem.setData(self, **kwargs)
        #elf.blockUpdate(False)
        self.sigPlotChanged.emit(self)
        
        print(f"self.pos: {self.pos}")
        
        

    def clipDataFromVRange(self, view_range, x, y, z):
        print(f"view_range: {view_range}")
        clipped_dataset=np.column_stack((x, y, z))
        for ax in range(3):
            mask = (view_range[ax][0] <= clipped_dataset[:, ax]) & (clipped_dataset[:, ax] <= view_range[ax][1])
            clipped_dataset = clipped_dataset[mask]
        return clipped_dataset[:, 0], clipped_dataset[:, 1], clipped_dataset[:,2]





from __future__ import annotations
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING



from ...gl_3D_graphing.graphics_items.GL3DAxisItem import GL3DAxisItem
import traceback

class PyVolGL3DAxisItem(GL3DAxisItem):
    def __init__(self, *args, **kwargs):
        self.tick_engine=kwargs.pop("tick_engine", None)
        super().__init__(*args, **kwargs)

    def tickStrings(self, values, scale, spacing):
        if not self.tick_engine is None:
            return self.tick_engine(values)
        else:
            return super().tickStrings(values, scale, spacing)
    
    def switch_axis(self, *args, **kwargs): 
        pass
    


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from instruments.instruments import Option
    from instruments.utils import InstrumentManager
    from ...quantities.engines import TickEngineManager
    from ...data_classes.classes import VolatilityData, Slice
    from ...workers import SurfaceCalibration

import pyqtgraph as pg
from PySide6 import QtWidgets
from typing import List
import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets
from ... import misc_widgets
from .. import subplot, custom_widgets
from ..volatility_surface.axis_widgets import CustomAxisItem
from typing import Dict
from pyvol_terminal.data_classes import builders as builders_data_classes
from ..abstract_classes import ABCMainViewQSplitter
from . import settings



class CustomPlotitem(subplot.SubPlot):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.plot_items={}
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.getViewBox().disableAutoRange(0)
        self.getViewBox().disableAutoRange(1)
        
        self._plot_items: List[pg.PlotDataItem] = []
        self.x_buffer = 0.1
        self.y_buffer = 0.1

    def add_legend(self, legend):
        self.legend = legend
    
    def addPlots(self, px_type, value, line, scatter):
        self.plot_items[px_type]["line"].update({value : line})
        self.plot_items[px_type]["scatter"].update({value : scatter})
        super().addItem(line)
        super().addItem(scatter)
        line.sigPlotChanged.emit(self.calculate_limits)
            
    def removeItem(self, item):
        if isinstance(item, pg.PlotDataItem) and item in self._plot_items:
            try:
                item.sigPlotChanged.disconnect(self.calculate_limits)
            except (RuntimeError, TypeError):
                pass 
            self._plot_items.remove(item)
        result = super().removeItem(item)
        self.calculate_limits()
    
    def _set_limits(self, x_min, x_max, y_min, y_max):
        if x_min == x_max:
            x_offset = 0.5 
        else:
            x_offset = (x_max - x_min) * self.x_buffer        
        if y_min == y_max:
            y_offset = 0.5
        else:
            y_offset = (y_max - y_min) * self.y_buffer
        self.getViewBox().setXRange(x_min - x_offset, x_max + x_offset)
        self.getViewBox().setYRange(y_min - y_offset, y_max + y_offset)    
        self.update()
        
    def calculate_limits(self):
        x_data = []
        y_data = []

        for plot_item in self._plot_items:
            x_min, x_max, y_min, y_max = plot_item.get_limits()
            x_data.append(x_min)
            x_data.append(x_max)
            y_data.append(y_min)
            y_data.append(y_max)
        if any(~np.isnan(x_data)):
            self.x_min, self.x_max = np.nanmin(x_data), np.nanmax(x_data)
        else:
            self.x_min, self.x_max = np.nan, np.nan
        if any(~np.isnan(y_data)):
            self.y_min, self.y_max = np.nanmin(y_data), np.nanmax(y_data)
        else:
            self.y_min, self.y_max = np.nan, np.nan
        
        if (not np.isnan(self.x_min)
            and not np.isnan(self.x_max)
            and not np.isnan(self.y_min)
            and not np.isnan(self.y_max)):
            if len(self._plot_items) == 1:
                self._set_limits(x_min, x_max, y_min, y_max)
            else:
                x_range, y_range = self.getPlotItem().getViewBox().viewRange()
                if (x_min < x_range[0] or x_max > x_range[1] or y_min < y_range[0] or y_max > y_range[1]):
                    self._set_limits(x_min, x_max, y_min, y_max)
        

class MainView(ABCMainViewQSplitter):
    def __init__(self,
                 instrument_manager: InstrumentManager,
                 interpolation_config,
                 tick_engine_manager: TickEngineManager,
                 slice_direction: str,
                 slice_type: str,
                 **kwargs 
                ):
        super().__init__(QtCore.Qt.Horizontal)
        self.volatility_data_container: Dict[str, VolatilityData] = kwargs.get("volatility_data_container", None)
        self.displayed_domains=[]
        self.container_scatter=[]
        self.container_scatter = {}
        self.container_line = {}
        self.displayed_domains=[]
        self.displayed_slices=[]
        self.diplayed_ptypes=[]
        self.slice_direction=slice_direction
        self.slice_type=slice_type
        self.tick_engine_manager=tick_engine_manager
        

        
        
        tick_engine_x = getattr(self.tick_engine_manager, f"{slice_direction}_engine")
        self.x_axis_item = CustomAxisItem(axis_direction="x",
                                        tick_engine=tick_engine_x,
                                        orientation="bottom",
                                        ) 
        self.y_axis_item = CustomAxisItem(axis_direction="x",
                                        tick_engine=self.tick_engine_manager.z_engine,
                                        orientation="right"
                                        )
                                                    
        
        axisItems = {"right" : self.y_axis_item, "bottom" : self.x_axis_item}
        
        self.plot = pg.PlotItem(axisItems=axisItems)
        self.plot.setSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Expanding)
                
        self.plot.addLegend()
        self.init_layout(tick_engine_x)


    def init_layout(self, tick_engine):
        
        container_widget = QtWidgets.QWidget()
        container_layout = QtWidgets.QGridLayout(container_widget)
        scroll = QtWidgets.QScrollArea()
        scroll.setWidgetResizable(True)
        toggle_widget = settings.ScrollableToggleButtons(self, getattr(self.domain, f"{self.slice_direction}_vect"), tick_engine)
        scroll.setWidget(toggle_widget)
        
        
        
        
        plot_widget = pg.PlotWidget(plotItem=self.plot)
        plot_widget.setLayout(container_layout)
        
        
        
     #   main_view_layout.addWidget(plot_widget, 0, 0)
        
        
     #   main_view_layout.addWidget(self.legend, 0, 0, alignment=QtCore.Qt.AlignTop | QtCore.Qt.AlignRight)
        
        self.addWidget(scroll)
       # self.addWidget(blank_surface_widget)
        self.addWidget(plot_widget)
        self.setStretchFactor(0, 5) 
        self.setStretchFactor(1, 1) 
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
    
    def addPlots(self, px_type):
        for value in self.displayed_domains:
            self.create_plot_objects(px_type, value)
    
    def addPlot(self, px_type, value):
        self.create_plot_objects(px_type, value)
    
    def create_plot_objects(self, px_type, value):
        self.data_container.add_slice(px_type, value)
        slice = self.data_container.slice_container[px_type][value]
        scatter = CustomPlotDataItem(x=slice.x, y=slice.y)
        line = CustomPlotDataItem(x=slice.xi, y=slice.yi)
        self.container_scatter[px_type][value] = scatter
        self.container_line[px_type][value] = line
    #    self.plot_view.addPlots(line)
        self.plot.addPlots(px_type, value, line, scatter)
        line.show()

    
    def _internal_update_view(self, calibrated_surface_container: Dict[str, SurfaceCalibration]):
        self.update_dataclasses(calibrated_surface_container)
        
    
    def update_dataclasses(self, calibrated_surface_container: Dict[str, SurfaceCalibration]):
        for px_type in self.diplayed_ptypes:
            surface_engine = calibrated_surface_container[px_type]
            volatility_data = self.volatility_data_container[px_type]
            dataclass = volatility_data.get_dataclass(self.slice_type)
            for slice_value in self.displayed_slices:
                pass



class CustomPlotWidget(pg.PlotWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.getPlotItem().getViewBox().disableAutoRange(0)
        self.getPlotItem().getViewBox().disableAutoRange(1)
        self._plot_items: List[pg.PlotDataItem] = []
        self.x_buffer = 0.1
        self.y_buffer = 0.1

    def add_curve(self, item):
        self.getPlotItem().addItem(item)
        self._plot_items.append(item)
        self.calculate_limits()  
        item.sigPlotChanged.emit(self.calculate_limits)
            
    def removeItem(self, item):
        if isinstance(item, pg.PlotDataItem) and item in self._plot_items:
            try:
                item.sigPlotChanged.disconnect(self.calculate_limits)
            except (RuntimeError, TypeError):
                pass 
            self._plot_items.remove(item)
        result = super().removeItem(item)
        self.calculate_limits()
    
    def _set_limits(self, x_min, x_max, y_min, y_max):
        if x_min == x_max:
            x_offset = 0.5 
        else:
            x_offset = (x_max - x_min) * self.x_buffer        
        if y_min == y_max:
            y_offset = 0.5
        else:
            y_offset = (y_max - y_min) * self.y_buffer
        self.getPlotItem().getViewBox().setXRange(x_min - x_offset, x_max + x_offset)
        self.getPlotItem().getViewBox().setYRange(y_min - y_offset, y_max + y_offset)    
        self.update()
        
    def calculate_limits(self):
        x_data = []
        y_data = []

        for plot_item in self._plot_items:
            x_min, x_max, y_min, y_max = plot_item.get_limits()
            x_data.append(x_min)
            x_data.append(x_max)
            y_data.append(y_min)
            y_data.append(y_max)
        if any(~np.isnan(x_data)):
            self.x_min, self.x_max = np.nanmin(x_data), np.nanmax(x_data)
        else:
            self.x_min, self.x_max = np.nan, np.nan
        if any(~np.isnan(y_data)):
            self.y_min, self.y_max = np.nanmin(y_data), np.nanmax(y_data)
        else:
            self.y_min, self.y_max = np.nan, np.nan
        
        if (not np.isnan(self.x_min)
            and not np.isnan(self.x_max)
            and not np.isnan(self.y_min)
            and not np.isnan(self.y_max)):
            if len(self._plot_items) == 1:
                self._set_limits(x_min, x_max, y_min, y_max)
            else:
                x_range, y_range = self.getPlotItem().getViewBox().viewRange()
                if (x_min < x_range[0] or x_max > x_range[1] or y_min < y_range[0] or y_max > y_range[1]):
                    self._set_limits(x_min, x_max, y_min, y_max)
        
class CustomScatterPlotItem(pg.ScatterPlotItem):
    pen_map: Dict[str, pg.mkPen]
    def __init__(self, px_type, *args, **kwargs):
        kwargs["symbol"] = "x"
        kwargs["size"]=10
        kwargs["pxMode"]=True
        self.px_type = px_type
        super().__init__(*args, **kwargs)



class CustomPlotDataItem(pg.PlotDataItem):
    pen_map: Dict[str, pg.mkPen]
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._calculate_limits(*self.getData())
    
    def _calculate_limits(self, x, y):
        if any(~np.isnan(x)):
            self.x_min, self.x_max = np.min(x), np.max(x)
        else:
            self.x_min, self.x_max = np.nan, np.nan
        if any(~np.isnan(y)):
            self.y_min, self.y_max = np.min(y), np.max(y)
        else:
            self.y_min, self.y_max = np.nan, np.nan
    
    def get_limits(self):
        return self.x_min, self.x_max, self.y_min, self.y_max
    
    def get_xlim(self):
        return self.x_min, self.x_max
    
    def get_ylim(self):
        return self.y_min, self.y_max

    def setData(self, **kwargs):
        self._calculate_limits(kwargs["x"], kwargs["y"])
        super().setData(**kwargs)
        
       


from __future__ import annotations
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from quantities.engines import TickEngine
    from pyvol_terminal.interfaces.volatility_surface.pyvol_GL3DViewWidget import SurfaceViewWidget
    
from PySide6 import QtGui, QtCore, QtWidgets
import pyqtgraph.opengl as gl
import numpy as np
from pyqtgraph.opengl import GLTextItem
from OpenGL import GL as opengl
from OpenGL import GLU
from PySide6 import QtGui
import math
import pyqtgraph as pg
from PySide6.QtGui import QColor, QPixmap, QPainter, QBrush
import warnings

import warnings
warnings.filterwarnings(
    "ignore",
    message=r"invalid value encountered in cast",
    category=RuntimeWarning  # or UserWarning if that’s what you see
)
    
class CustomColorMap(pg.ColorMap):
    def __init__(self, colourmap_style: str):
        pos = np.linspace(0, 1, 500)          
        try:
            colourmap = pg.colormap.get(colourmap_style)
        except:
            print(f"{colourmap_style} is not in pyqtgraph, using default inferno")
            colourmap = pg.colormap.get("inferno")

        colors = colourmap.map(pos, mode='byte')        
        super().__init__(pos=pos, color=colors, mode='byte')


class ColorSquare(QtWidgets.QLabel):
    def __init__(self, color, size=50):
        super().__init__()
        self.color = color
        self.color_width = size
        self.color_height = size /2
        self.setFixedSize(self.color_width, self.color_height)
        self.update_color()

    def update_color(self):
        pixmap = QPixmap(self.color_width, self.color_height)
        pixmap.fill(QtCore.Qt.transparent)
        painter = QPainter(pixmap)
        if isinstance(self.color, tuple):
            r, g, b, a = [int(x * 255) for x in self.color]
            q_color = QColor(r, g, b, a)
            
        elif isinstance(self.color, str):
            
            q_color = QColor(self.color)
        else:
            q_color =None

        if not q_color is None:
            painter.setBrush(QBrush(q_color))
        painter.drawRect(0, 0, self.color_width, self.color_height)
        painter.end()
        self.setPixmap(pixmap)
        self.setStyleSheet("border: 2px solid black;")


class Legend(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        self.setStyleSheet("background: transparent; color: white;")
        self.layout = QtWidgets.QVBoxLayout(self)
        self.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.legend_items={}
        self.legend_count={}
        
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Expanding
                           )

    def remove_legend_item(self, name):
        self.legend_count[name]-=1
        if self.legend_count[name]==0:
            inner_widget=self.legend_items[name]
            self.layout.removeWidget(inner_widget)
                    
    def add_legend_item(self, name, colour):
        if not name in self.legend_count:
            self.legend_count[name]=1
        else:
            self.legend_count[name]+=1   
        if self.legend_count[name] > 1:
            return
        
        hbox_layout = QtWidgets.QHBoxLayout()
        hbox_layout.setAlignment(QtCore.Qt.AlignLeft)
        hbox_layout.setContentsMargins(0, 0, 0, 0)
        hbox_layout.setSpacing(5)
        
        square_colour = ColorSquare(colour)
        hbox_layout.addWidget(square_colour)

        square_label = QtWidgets.QLabel(name)
        square_label.setStyleSheet("border: 1px solid white; padding: 3px; background-color: white; color: black")
        hbox_layout.addWidget(square_label)

        inner_widget = QtWidgets.QWidget()
        inner_widget.setLayout(hbox_layout)

        self.legend_items[name] = inner_widget        
        self.layout.addWidget(inner_widget)
        self.adjustSize()
        self.update()


import numpy as np
from dataclasses import dataclass
import time
from datetime import datetime
import numpy as np
import pandas as pd
from abc import ABC, abstractmethod

class CustomDatetime:
    @staticmethod
    def fromtimestamp(ts):
        if ts != ts: 
            return str(ts)
        return datetime.fromtimestamp(ts)


def cleanup_dataframe(df, df_type):
    if df_type == "options":
        df = df.sort_values(by=["expiry", "strike", "flag"], ascending=[True, True, False])
        df.reset_index(drop=True, inplace=True)
        
        for idx, opt_type in zip(df.index, df["flag"]):
            if opt_type.lower() == "call":
                df.loc[idx, "flag"] = "c"
            elif opt_type.lower() == "put":
                df.loc[idx, "flag"] = "p"

    elif df_type == "futures" or df_type == "options":
        df = df.sort_values(by=["expiry"], ascending=[True])
    
    numeric_columns = ["strike", "expiry", "bid", "ask", "mid"]
    for col in numeric_columns:
        if col in df.columns:
            df[col] = pd.to_numeric(df[col], errors='coerce')
    return df

class AbstractValidPriceChecker(ABC):
    
    @abstractmethod
    def __init__(self, *args):
        ...

    @abstractmethod
    def check(self, *args):
        ...

class ValidPriceChecker(AbstractValidPriceChecker):
    def __init__(self, spread_minimum):
        self.spread_minimum = spread_minimum  
        
    def check(self, bid, ask):
        if (np.isnan(bid)
            or np.isnan(ask)
            or bid==0
            or ask==0
            or 100 * (ask - bid) / (0.5 * (bid + ask)) > self.spread_minimum
            ):
            return False
        else:
            return True

def filter_nans_2D(x, y):
    return x[~np.isnan(y)], y[~np.isnan(y)]

def filter_nans_on_z(x, y, z):
    return x[~np.isnan(z)], y[~np.isnan(z)], z[~np.isnan(z)]

def filter_nans_on_z_OTM(x, y, OTM, z):
    return x[~np.isnan(z)], y[~np.isnan(z)], OTM[~np.isnan(z)], z[~np.isnan(z)]


def check_same_nan_structure(arr1, arr2):
    mask1 = np.isfinite(arr1)  
    mask2 = np.isfinite(arr2)  
    return np.array_equal(mask1, mask2)

@dataclass(frozen=True, slots=True)
class ws_pyvol_keymap:
    ticker: str
    bid: str
    ask: str
    timestamp: str

def create_websocket_pyvol_keymapper(ticker_key, bid_key, ask_key, timestamp_key) -> ws_pyvol_keymap:
    return ws_pyvol_keymap(ticker_key,
                           bid_key,
                           ask_key,
                           timestamp_key
                           )

def convert_unix_maturity_to_years(unix_maturity):
    return (unix_maturity - time.time()) / 3600 / 24 / 365

def convert_to_yte(timestamp, unix_maturity):
    return (unix_maturity - timestamp) / 3600 / 24 / 365

def calculate_xy_lines(X, Y, Z, x_fixed, y_fixed, const_axis):
    if const_axis:
        y_index = np.argmin(np.abs(Y - y_fixed))  
        line = np.column_stack((X, [y_fixed]*X.size, Z[:,y_index]))     
    else:
        x_index = np.argmin(np.abs(X - x_fixed)) 
        line = np.column_stack(([x_fixed] * Y.size, Y, Z[x_index,: ]))  
    return line


class BiDict:
    def __init__(self):
        self._type1_to_type2 = {}  # Maps type1 keys to type2 values
        self._type2_to_type1 = {}  # Maps type2 keys to type1 values

    def add(self, type1, type2):
        """Add a bidirectional mapping between type1 and type2."""
        self._type1_to_type2[type1] = type2
        self._type2_to_type1[type2] = type1

    def get_type2(self, type1):
        """Get the type2 value associated with a type1 key."""
        return self._type1_to_type2.get(type1)

    def get_type1(self, type2):
        """Get the type1 value associated with a type2 key."""
        return self._type2_to_type1.get(type2)

    def remove_type1(self, type1):
        """Remove a mapping by type1 key."""
        if type1 in self._type1_to_type2:
            type2 = self._type1_to_type2.pop(type1)
            del self._type2_to_type1[type2]

    def remove_type2(self, type2):
        """Remove a mapping by type2 key."""
        if type2 in self._type2_to_type1:
            type1 = self._type2_to_type1.pop(type2)
            del self._type1_to_type2[type1]

    @property
    def type1(self):
        """Return an iterable view of type1 keys."""
        return self._type1_to_type2.keys()

    @property
    def type2(self):
        """Return an iterable view of type2 keys."""
        return self._type2_to_type1.keys()

    def iter_type1(self):
        """Iterate over type1 keys and their corresponding type2 values."""
        return self._type1_to_type2.items()

    def iter_type2(self):
        """Iterate over type2 keys and their corresponding type1 values."""
        return self._type2_to_type1.items()

    def __contains__(self, key):
        """Check if a key exists in either type1 or type2."""
        return key in self._type1_to_type2 or key in self._type2_to_type1

    def __len__(self):
        """Return the number of mappings."""
        return len(self._type1_to_type2)

    def __repr__(self):
        """Return a string representation of the BiDict."""
        return f"BiDict(type1_to_type2={self._type1_to_type2}, type2_to_type1={self._type2_to_type1})"


#from pyqtgraph.Qt import QtCore, QtGui, QtWidgets
from PySide6 import QtCore, QtGui, QtWidgets
from pyqtgraph import opengl
from OpenGL import GL
from pyqtgraph import Transform3D
import weakref
import numpy as np
from pyqtgraph.Qt import QtCore
from pyvol_terminal.gl_3D_graphing.graphics_items.GLGraphicsObject import GLGraphicsObject
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DViewBox import GLChildGroup
from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DItemGroup import GL3DItemGroup
import weakref
from pyqtgraph.opengl import GLGraphicsItem
import weakref
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets
import operator
from functools import reduce
from abc import ABCMeta, abstractmethod
import numpy as np
import warnings
from pyqtgraph.Qt import QT_LIB
from pyqtgraph import Transform3D


class WeakList(object):

    def __init__(self):
        self._items = []

    def append(self, obj):
        #Add backwards to iterate backwards (to make iterating more efficient on removal).
        self._items.insert(0, weakref.ref(obj))

    def __iter__(self):
        i = len(self._items)-1
        while i >= 0:
            ref = self._items[i]
            d = ref()
            if d is None:
                del self._items[i]
            else:
                yield d
            i -= 1


class AbstractGLGraphicsItem(GLGraphicsItem.GLGraphicsItem):
    
    def __init__(self, parentItem=None, **kwargs):
        self._blockUpdate=None
        self._viewBox=None
        print("AbstractGLGraphicsItem")
        print(f"parentItem: {parentItem}")
        
        super().__init__(parentItem=parentItem)

    @property
    def _parent(self):
        """Access to __parent"""
        return self._GLGraphicsItem__parent
    
    @_parent.setter
    def _parent(self, value):
        """Set __parent"""
        self._GLGraphicsItem__parent = value
    
    @property
    def _view(self):
        """Access to __view"""
        return self._GLGraphicsItem__view
    
    @_view.setter
    def _view(self, value):
        """Set __view"""
        self._GLGraphicsItem__view = value
    
    @property
    def _children(self):
        """Access to __children"""
        return self._GLGraphicsItem__children
    
    @property
    def _transform(self):
        """Access to __transform"""
        return self._GLGraphicsItem__transform
    
    @_transform.setter
    def _transform(self, value):
        """Set __transform"""
        self._GLGraphicsItem__transform = value
    
    @property
    def _visible(self):
        """Access to __visible"""
        return self._GLGraphicsItem__visible
    
    @_visible.setter
    def _visible(self, value):
        """Set __visible"""
        self._GLGraphicsItem__visible = value
    
    @property
    def _initialized(self):
        """Access to __initialized"""
        return self._GLGraphicsItem__initialized
    
    @_initialized.setter
    def _initialized(self, value):
        """Set __initialized"""
        self._GLGraphicsItem__initialized = value
    
    @property
    def _glOpts(self):
        """Access to __glOpts"""
        return self._GLGraphicsItem__glOpts
    
    @_glOpts.setter
    def _glOpts(self, value):
        """Set __glOpts"""
        self._GLGraphicsItem__glOpts = value
    
    @property
    def _depthValue(self):
        """Access to __depthValue"""
        return self._GLGraphicsItem__depthValue
    
    @_depthValue.setter
    def _depthValue(self, value):
        """Set __depthValue"""
        self._GLGraphicsItem__depthValue = value
    
class GLGraphicsObject(AbstractGLGraphicsItem, QtWidgets.QGraphicsObject): 
    _qtBaseClass = QtWidgets.QGraphicsObject
    def __init__(self, parentItem=None):
        self.__inform_view_on_changes = True
        QtWidgets.QGraphicsObject.__init__(self)
        
        self.setFlag(self.GraphicsItemFlag.ItemSendsGeometryChanges)

        
    def itemChange(self, change, value):
        ret = super().itemChange(change, value)
        if change in [self.GraphicsItemChange.ItemParentHasChanged, self.GraphicsItemChange.ItemSceneHasChanged]:
            if self.__class__.__dict__.get('parentChanged') is not None:
                # user's GraphicsObject subclass has a parentChanged() method
                warnings.warn(
                    "parentChanged() is deprecated and will be removed in the future. "
                    "Use changeParent() instead.",
                    DeprecationWarning, stacklevel=2
                )
                if QT_LIB == 'PySide6' and QtCore.__version_info__ == (6, 2, 2):
                    # workaround PySide6 6.2.2 issue https://bugreports.qt.io/browse/PYSIDE-1730
                    # note that the bug exists also in PySide6 6.2.2.1 / Qt 6.2.2
                    getattr(self.__class__, 'parentChanged')(self)
                else:
                    self.parentChanged()
            else:
                print("itemchange")
                self.changeParent()
        try:
            inform_view_on_change = self.__inform_view_on_changes
        except AttributeError:
            # It's possible that the attribute was already collected when the itemChange happened
            # (if it was triggered during the gc of the object).
            pass
        else:
            if inform_view_on_change and change in [self.GraphicsItemChange.ItemPositionHasChanged, self.GraphicsItemChange.ItemTransformHasChanged]:
                self.informViewBoundsChanged()
            
        return ret

    def parentItem(self):
        return self.__parent
    
class GL3DItemGroup(GLGraphicsObject):
    """
    Replacement for QGraphicsItemGroup
    """
    
    def __init__(self, *args):
        GLGraphicsObject.__init__(self, *args)
        self.setFlag(self.GraphicsItemFlag.ItemHasNoContents)
    
    def boundingRect(self):
        return QtCore.QRectF()
        
    def paint(self, *args):
        pass
    
    def addItem(self, item):
        item.setParentItem(self)


class GLChildGroup(GL3DItemGroup):

    def __init__(self, parentItem):
        GL3DItemGroup.__init__(self, parentItem)

        # Used as callback to inform ViewBox when items are added/removed from
        # the group.
        # Note 1: We would prefer to override itemChange directly on the
        #         ViewBox, but this causes crashes on PySide.
        # Note 2: We might also like to use a signal rather than this callback
        #         mechanism, but this causes a different PySide crash.
        self.itemsChangedListeners = WeakList()

        # exempt from telling view when transform changes
        self._GraphicsObject__inform_view_on_change = False

    def itemChange(self, change, value):
        ret = GL3DItemGroup.itemChange(self, change, value)
        if change in [
            self.GraphicsItemChange.ItemChildAddedChange,
            self.GraphicsItemChange.ItemChildRemovedChange,
        ]:
            try:
                itemsChangedListeners = self.itemsChangedListeners
            except AttributeError:
                # It's possible that the attribute was already collected when the itemChange happened
                # (if it was triggered during the gc of the object).
                pass
            else:
                for listener in itemsChangedListeners:
                    listener.itemsChanged()
        return ret
    
    def linkViewBox(self, vb):
        print
        self._viewBox=vb
        if self.parentItem() != vb:
            self.setParentItem(vb)
        self.update()


class CustomItem(AbstractGLGraphicsItem, QtWidgets.QGraphicsItem):
    def __init__(self):
        QtWidgets.QGraphicsItem.__init__(self, None)
        
def main():
    
    app = QtWidgets.QApplication()
    win = QtWidgets.QMainWindow()
    item1 = CustomItem()
    item2 = CustomItem()
    item1.setParentItem
    #child_group = GLChildGroup(parentItem=item)
    #child_group.setParentItem()
 #   print(gitem.parentItem())
    win.show()
    app.exec()

if __name__ == "__main__":
    main()



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from instruments.instruments import Option
    from instruments.utils import InstrumentManager
    
import pyqtgraph.opengl as gl
import numpy as np
from PySide6 import QtGui, QtCore
import uuid

class GlPlotItemMixin:
    sigPlotChanged = QtCore.Signal(object)
    
    def __init__(self, px_type, type, *args, **kwargs):
        self.px_type=px_type
        self.item_type=type
        self._internal_id=uuid.uuid4()
        self.color=kwargs["color"]
        self.original_setData: Callable=None
        self.dataset = PlotDataset(real_all)
        
        if isinstance(kwargs["color"], str):
            qcolor =  QtGui.QColor(kwargs["color"])
            normalised_rgba = (qcolor.redF(), qcolor.greenF(), qcolor.blueF(), qcolor.alphaF())
            kwargs["color"] = normalised_rgba
        super().__init__(*args, **kwargs)

    def setData(self, *args, **kwargs):
        super().setData(*args, **kwargs)
        self.sigPlotChanged.emit(self)

    def id(self):
        return self._internal_id
    
class glSurface(GlPlotItemMixin, gl.GLSurfacePlotItem):
    def __init__(self, px_type, colormap, color, *args, **kwargs):
        kwargs["shader"] = "shaded"
        self.colormap=colormap
        self.color=color
        self.init=True
        kwargs["color"]=color
        kwargs["colors"]=colormap.map(kwargs["z"])
        self.valid_values=False
        super().__init__(px_type, "surface", *args, **kwargs)
        self.setGLOptions('opaque')
    
    def setData(self, *args, **kwargs):
        if len(args) >= 4 or "colors" in kwargs:
            return super().setData(*args, **kwargs)
        else:
            if len(args) >= 3:
                z_values = args[2]             
            else:
                z_values = kwargs.get("z")       
            if z_values is not None:
                kwargs["colors"] = self.colormap.map(z_values)
            super().setData(*args, **kwargs)     
            self.valid_values=~np.isnan(self._z).all()
            
    def getValues(self,):
        return self._x, self._y, self._z
    
class glScatter(GlPlotItemMixin, gl.GLScatterPlotItem):
    def __init__(self, px_type, *args, **kwargs):
        kwargs["size"] = 10
        super().__init__(px_type, "scatter", *args, **kwargs)
        
    def getValues(self):
        return self.pos[:,0], self.pos[:,1], self.pos[:,2]



import numpy as np
import pyqtgraph as pg
import pyqtgraph.opengl as gl
from PySide6 import QtWidgets
import sys

def main():
    app = QtWidgets.QApplication(sys.argv)

    # Create the main OpenGL view widget
    w = gl.GLViewWidget()
    w.setWindowTitle('GLSurfacePlotItem Example')
    w.setCameraPosition(distance=40)
    w.show()

    # Create grid
    g = gl.GLGridItem()
    g.scale(2, 2, 1)
    w.addItem(g)

    # Generate a 2D surface (z = sin(r))
    x = np.linspace(-10, 10, 50)
    y = np.linspace(-10, 10, 50)
    X, Y = np.meshgrid(x, y)
    r = np.sqrt(X**2 + Y**2)
    z = np.sin(r)
    
    z = np.nan * z
    # Create surface plot
    p = gl.GLSurfacePlotItem(x=x, y=y, z=z, shader='shaded')
    p.translate(0, 0, 0)  # Optional: move the surface
    w.addItem(p)

    sys.exit(app.exec())

if __name__ == '__main__':
    main()



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from ..data_classes.classes import OptionChain
    from pyvol_terminal.engines.interpolation_engines import Abstract3DInterpolator

import numpy as np
from abc import ABC, abstractmethod
from scipy import interpolate
from ..utils import filter_nans_on_z, filter_nans_on_z_OTM
import math
import time


class AbstractSurfaceEngine(ABC):
    def __init__(self, n_tolerance=5, *args, **kwargs):
        self.n_tolerance=n_tolerance
        self._calibrated=False
        
    def valid_calibration(self):
        return self._calibrated
    
    @abstractmethod
    def calibrate(self, *args, **kwargs) -> None:
        ...
        
    @abstractmethod
    def evaluate(self,
                 xi: np.ndarray,
                 yi: np.ndarray,
                 *args: Tuple[Any],
                 **kwargs: Dict[str, Any]
                 ) -> np.ndarray:
        ...
        

class ATMWeightedSurfaceEngine(AbstractSurfaceEngine):
    def __init__(self, *args, **kwargs):
        #super(ATMWeightedSurfaceEngine, self).__init__(*args, **kwargs)
        super().__init__(*args, **kwargs)
        self.kx, self.ky = 3, 3
    
    def calibrate(self,
                  x: np.ndarray,
                  y: np.ndarray,
                  z: np.ndarray,
                  OTM: np.ndarray,
                  ) -> None:
        x, y, OTM, z = filter_nans_on_z_OTM(x, y, OTM, z)
        if x.size < self.n_tolerance:
            self._calibrated=False
            return
        else:
            m = x.size
            
            if m >= (self.kx+1) * (self.ky+1):
                
                y = (y - time.time()) / 3600 / 24 / 365
                z *=z
                
                self.tck = interpolate.bisplrep(x, y, z, kx=self.kx, ky=self.ky, s=0.5 * (m-math.sqrt(2*m) + m+math.sqrt(2*m))) 
                self._calibrated=True
            else:
                self._calibrated=False
    

    
    def evaluate(self,
                 xi: np.ndarray,
                 yi: np.ndarray,
                 *args: Tuple[Any],
                 **kwargs: Dict[str, Any]
                 ) -> np.ndarray:
        if not self._calibrated:
            return 
        x_vals = np.unique(xi)
        y_vals = np.unique(yi)
        y_vals = np.unique(yi)
        y_vals = (y_vals - time.time()) / 3600 / 24 / 365
        variance = interpolate.bisplev(x_vals, y_vals, self.tck, *args, **kwargs) 
        return np.sqrt(variance)



from __future__ import annotations 
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DViewBox import GL3DViewBox

from pyqtgraph import opengl
from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
from OpenGL import GL
import sys
import cProfile
import pstats



class Window(QtWidgets.QMainWindow):
    def __init__(self, *args, app=None, **kwargs):
        super().__init__(*args, **kwargs)   
    
        self.gl_view = opengl.GLViewWidget()
        
        self.setCentralWidget(self.gl_view)
        self.count=0
        self.m=50
        self.world_m = 4
        self.gl_view.opts.update({"azimuth" : -50,
                          "distance" : 10,
                          "center" : QtGui.QVector3D(0, 1, 0)}
                         )
        
        self.gl_view.update()
        self.linedata_even = np.array([[0., 1., 0.,], [1., 1., 0.]] * self.m)
        self.linedata_odd = np.array([[0., 1., 1.,], [1., 1., 1.]] * self.m)
        self.linedata_even[:, 1] = self.linedata_even[:, 1] * np.repeat(np.linspace(0, self.world_m, self.m), 2)
        self.linedata_odd[:, 1] = self.linedata_odd[:, 1] * np.repeat(np.linspace(0, self.world_m, self.m), 2)
        pos = self.linedata_odd
        self.lines1 = opengl.GLLinePlotItem(pos=self.linedata_odd,
                                           color=(255, 255, 255, 150),
                                           glOptions="translucent",
                                           mode="lines"
                                           )
        self.gl_view.addItem(self.lines1)
        line_data2 = pos.copy()
        line_data2[:, 1] += line_data2[0, 1] - line_data2[1, 1]
        line_data2[:, 1] = line_data2[:, 1] + pos[:, 1]
        self.lines2 = opengl.GLLinePlotItem(pos=line_data2,
                                            color=(255, 255, 255, 150),
                                            glOptions="translucent",
                                            mode="lines"
                                            )

        
        self.gl_view.addItem(self.lines2)

        
        self.showMaximized()
        
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_plots)
        


    def update_plots(self):
        self.count+=1
        if self.count % 2 == 0:
            pos = self.linedata_even
        else:
            pos = self.linedata_odd

        self.lines1.setData(pos=pos)
        self.lines1.paint()

        line_data2 = pos.copy()
        line_data2[:, 1] += line_data2[0, 1] - line_data2[1, 1]
        line_data2[:, 1] = line_data2[:, 1] + pos[:, 1]
        self.lines2.setData(pos=line_data2)


if __name__ == "__main__":
    

    profiler = cProfile.Profile()
    profiler.enable()
    
    app = QtWidgets.QApplication(sys.argv)
    win = Window()
    win.timer.start(100)
    app.exec()
    
    profiler.disable()


    with open(f'profit_{win.m}.txt', 'w') as f:
        ps = pstats.Stats(profiler, stream=f)
        ps.sort_stats('tottime')
        ps.print_stats()





from OpenGL.GL import *  # noqa
import numpy as np
import math
import importlib
from OpenGL import GL


from pyqtgraph.Qt import QtGui, QT_LIB
from OpenGL.GL import shaders
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem

QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")
__all__ = ['GLScatterPlotItem']


class CustomGLScatterPlotItem2(GLGraphicsItem):
    """Draws points at a list of 3D positions."""
    
    _shaderProgram = None

    def __init__(self, parentItem=None, **kwds):
        super().__init__()
        glopts = kwds.pop('glOptions', 'opaque')
        self.setGLOptions(glopts)
        self.vertices = None 
        if isinstance(kwds["color"], str):
            qcolor =  QtGui.QColor(kwds["color"])
            normalised_rgba = (qcolor.redF(), qcolor.greenF(), qcolor.blueF(), qcolor.alphaF())
            kwds["color"] = normalised_rgba
        self.color=normalised_rgba
        self.size = 10
        self.pxMode = True

        self.m_vbo_position = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.m_vbo_color = QtOpenGL.QOpenGLBuffer(QtOpenGL.QOpenGLBuffer.Type.VertexBuffer)
        self.vbos_uploaded = False

        self.setParentItem(parentItem)
        self.setData(**kwds)

    def setData(self, **kwds):
        args = ['pos', 'color', 'size']
        for k in kwds.keys():
            if k not in args:
                raise Exception(f'Invalid keyword argument: {k} (allowed arguments are {str(args)})')
            
        if 'pos' in kwds:
            pos = kwds.pop('pos')
            # Convert positions to triangles (each point becomes a small triangle)
            self.vertices = self._create_triangle_vertices(pos)
            
        if 'color' in kwds:
            color = kwds.pop('color')
            if isinstance(color, np.ndarray):
                color = np.ascontiguousarray(color, dtype=np.float32)
            self.color = color
        if 'size' in kwds:
            size = kwds.pop('size')
            if isinstance(size, np.ndarray):
                size = np.ascontiguousarray(size, dtype=np.float32)
            self.size = size

            
        self.pxMode = kwds.get('pxMode', self.pxMode)

        self.vbos_uploaded = False
        self.update()

    def _create_triangle_vertices(self, centers):
        vertices = []
        base_size = 0.01  # This empirically matches GLScatterPlotItem's size
        
        for center in centers:
            # Create a small triangle in screen space
            v1 = center + np.array([-base_size, -base_size, 0])
            v2 = center + np.array([base_size, -base_size, 0])
            v3 = center + np.array([0, base_size, 0])
            
            vertices.extend([v1, v2, v3])
            
        return np.ascontiguousarray(vertices, dtype=np.float32)
    
    def paint(self):
        if self.vertices is None:
            return

        self.setupGLState()
        
        mat_mvp = self.mvpMatrix()
        mat_mvp = np.array(mat_mvp.data(), dtype=np.float32)
        mat_modelview = self.modelViewMatrix()
        mat_modelview = np.array(mat_modelview.data(), dtype=np.float32)

        program = self.getShaderProgram()
        
        # Calculate scaling factor
        scale = 0
        if self.pxMode:
            view = self.view()
            if view is not None:
                scale = 2.0 * math.tan(math.radians(0.5 * view.opts["fov"])) / view.height()

        if not self.vbos_uploaded:
            self.upload_vbo(self.m_vbo_position, self.vertices)
            if isinstance(self.color, np.ndarray):
                # Repeat colors for each vertex in the triangle
                colors = np.repeat(self.color, len(self.vertices)//len(self.color), axis=0)
                self.upload_vbo(self.m_vbo_color, colors)
            self.vbos_uploaded = True

        enabled_locs = []

        if (loc := GL.glGetAttribLocation(program, "a_position")) != -1:
            self.m_vbo_position.bind()
            GL.glVertexAttribPointer(loc, 3, GL.GL_FLOAT, False, 0, None)
            self.m_vbo_position.release()
            enabled_locs.append(loc)

        if (loc := GL.glGetAttribLocation(program, "a_color")) != -1:
            if isinstance(self.color, np.ndarray):
                self.m_vbo_color.bind()
                GL.glVertexAttribPointer(loc, 4, GL.GL_FLOAT, False, 0, None)
                self.m_vbo_color.release()
                enabled_locs.append(loc)
            else:
                color = self.color
                if isinstance(color, QtGui.QColor):
                    color = color.getRgbF()
                GL.glVertexAttrib4f(loc, *color)

        for loc in enabled_locs:
            GL.glEnableVertexAttribArray(loc)

        with program:
            # Set uniforms
            loc = GL.glGetUniformLocation(program, "u_mvp")
            GL.glUniformMatrix4fv(loc, 1, False, mat_mvp)
            loc = GL.glGetUniformLocation(program, "u_modelview")
            GL.glUniformMatrix4fv(loc, 1, False, mat_modelview)
            loc = GL.glGetUniformLocation(program, "u_scale")
            GL.glUniform1f(loc, scale)
            
            GL.glDrawArrays(GL.GL_TRIANGLES, 0, len(self.vertices))
            
        for loc in enabled_locs:
            GL.glDisableVertexAttribArray(loc)
            
    def upload_vbo(self, vbo, arr):
        if arr is None:
            vbo.destroy()
            return
        if not vbo.isCreated():
            vbo.create()
        vbo.bind()
        vbo.allocate(arr, arr.nbytes)
        vbo.release()

    @staticmethod
    def getShaderProgram():
        klass = CustomGLScatterPlotItem2

        if klass._shaderProgram is not None:
            return klass._shaderProgram

        ctx = QtGui.QOpenGLContext.currentContext()
        fmt = ctx.format()

        if ctx.isOpenGLES():
            if fmt.version() >= (3, 0):
                glsl_version = "#version 300 es\n"
                sources = SHADER_CORE
            else:
                glsl_version = "#version 100\n"
                sources = SHADER_LEGACY
        else:
            if fmt.version() >= (3, 1):
                glsl_version = "#version 140\n"
                sources = SHADER_CORE
            else:
                glsl_version = "#version 120\n"
                sources = SHADER_LEGACY

        compiled = [shaders.compileShader([glsl_version, v], k) for k, v in sources.items()]
        program = shaders.compileProgram(*compiled)

        # bind generic vertex attrib 0 to "a_position" so that
        # vertex attrib 0 definitely gets enabled later.
        GL.glBindAttribLocation(program, 0, "a_position")
        GL.glLinkProgram(program)

        klass._shaderProgram = program
        return program

    def setupGLState(self):
        glEnable(GL_DEPTH_TEST)  # Crucial for proper depth sorting
        glDisable(GL_BLEND)      # Disable blending
        glDepthFunc(GL_LESS)     # Standard depth comparison
        glDisable(GL_ALPHA_TEST) # Not needed for opaque objects
        
def _is_compatibility_profile(context):
    # https://stackoverflow.com/questions/73745603/detect-the-opengl-context-profile-before-version-3-2
    sformat = context.format()
    profile = sformat.profile()

    # >= 3.2 has {Compatibility,Core}Profile
    # <= 3.1 is NoProfile

    if profile == sformat.OpenGLContextProfile.CompatibilityProfile:
        compat = True
    elif profile == sformat.OpenGLContextProfile.CoreProfile:
        compat = False
    else:
        compat = False
        version = sformat.version()

        if version <= (2, 1):
            compat = True
        elif version == (3, 0):
            if sformat.testOption(sformat.FormatOption.DeprecatedFunctions):
                compat = True
        elif version == (3, 1):
            if context.hasExtension(b'GL_ARB_compatibility'):
                compat = True

    return compat


## See:
##
##  http://stackoverflow.com/questions/9609423/applying-part-of-a-texture-sprite-sheet-texture-map-to-a-point-sprite-in-ios
##  http://stackoverflow.com/questions/3497068/textured-points-in-opengl-es-2-0
##
##
SHADER_LEGACY = {
    GL.GL_VERTEX_SHADER : """
        uniform mat4 u_mvp;
        uniform mat4 u_modelview;
        uniform float u_scale;
        attribute vec4 a_position;
        attribute vec4 a_color;
        varying vec4 v_color;

        void main() {
            gl_Position = u_mvp * a_position;
            v_color = a_color;
            
            if (u_scale != 0.0) {
                vec4 cpos = u_modelview * a_position;
                float dist = length(cpos.xyz);
                gl_PointSize = %f * self.size / (dist * u_scale);
            }   
        }
    """% (0.01 * 100),
    GL.GL_FRAGMENT_SHADER : """
        #ifdef GL_ES
        precision mediump float;
        #endif

        varying vec4 v_color;
        void main() {
            gl_FragColor = v_color;
        }
    """
}

SHADER_CORE = {
    GL.GL_VERTEX_SHADER : """
        uniform mat4 u_mvp;
        in vec4 a_position;
        in vec4 a_color;
        out vec4 v_color;

        void main() {
            gl_Position = u_mvp * a_position;
            v_color = a_color;
        }
    """,
    GL.GL_FRAGMENT_SHADER : """
        #ifdef GL_ES
        precision mediump float;
        #endif

        in vec4 v_color;
        out vec4 fragColor;
        void main() {
            fragColor = v_color;
        }
    """
}


#%%%
from functools import wraps

class GrandMother:
    wrapped_methods = "paint"
    def __init__(self, *args,):
        super().__init__()
        self._old_paint=None

    def __init_subclass__(cls, **kwargs):        
        super().__init_subclass__(**kwargs)


        for method_name in cls.__dict__:
            if method_name in cls.wrapped_methods:
                method = cls.__dict__[method_name].__name__
                print(f"method: {method}")
                @wraps(method)
                def wrapped(self, *args, **kwargs):
                    helper_method = getattr(GrandMother, f"_{method}Helper")
                    return helper_method(self, *args, **kwargs)
                
                setattr(cls, method_name, wrapped)
        
        
    def paint(self):
        print("grandmother")
        
    def _paintHelper(self):
        print(f"__paintHelper")
        print(self.__class__.paint())
        

class Parent3(GrandMother):
    def __init__(self):
        super().__init__()
    def paint(self):
        print("parent")

class Child(Parent3):
    def __init__(self):
        super().__init__()

    def paint(self):
        print("child")

a = Child()
a.paint()

#%%%

class grandParent:
    def __init__(self):
        print("gp before")
        super().__init__()


class Parent1(grandParent):
    def __init__(self):
        print("p1 before")
        super().__init__()
        #print("p1 after")
        
    def func(self):
        print("Parent1")



class Parent2:
    def __init__(self):
        print("p2 before")
        super().__init__()
        #print("p2 after")
    
    def func(self):
        print("Parent2")
        
        
class Child1(Parent1, Parent2): 
    def __init__(self):
        super().__init__()
        
Child1()
    
#%%
"""
class D:
    def __init__(self):
        print("D")
        super().__init__()

class B(D):
    def __init__(self):
        print("B")
        super().__init__()

class C:
    def __init__(self):
        print("C")
        super().__init__()

class A(B, C):  
    def __init__(self):
        print("A")
        super().__init__()

A()
#%%

from PySide6 import QtCore, QtWidgets, QtGui
import warnings
import abc
from PySide6.QtCore import QObject
from pyqtgraph import Transform3D
from pyvol_terminal.gl_3D_graphing.graphics_items import CustomGLGraphicsItem
from pyvol_terminal.gl_3D_graphing.graphics_items import GLGraphicsObject

class GLGroupItem(CustomGLGraphicsItem.CustomGLGraphicsItem):
    def __init__(self, parentItem=None):
        super().__init__(parentItem=parentItem)


a=GLGraphicsObject.GLGraphicsObject()
#a.rect()
print(a)
import sys
sys.exit()
"""

"""
from pyvol_terminal.gl_3D_graphing.graphics_items import GLGraphicsObject

class MyClass(GLGraphicsObject.GLGraphicsObject):
    def __init__(self, *args,):
        super().__init__()

MyClass()
from pyvol_terminal.gl_3D_graphing.graphics_items import AbstractGLPlotItem
from abc import ABCMeta
import abc
from PySide6 import QtCore, QtWidgets, QtGui

from pyqtgraph.opengl import GLGraphicsItem

class QABCMeta(ABCMeta, type(QtCore.QObject)):
    
    def __new__(mcls, name, bases, ns, **kw):
        cls = super().__new__(mcls, name, bases, ns, **kw)
        abc._abc_init(cls)
        return cls
    def __call__(cls, *args, **kw):
        if cls.__abstractmethods__:
            raise TypeError(f"Can't instantiate abstract class {cls.__name__} without an implementation for abstract methods {set(cls.__abstractmethods__)}")
        return super().__call__(*args, **kw)




class GraphicsItemMeta(ABCMeta, type(QtCore.QObject)):
    pass

class AbstractGLGraphicsItem(QtWidgets.QGraphicsItem, GLGraphicsItem.GLGraphicsItem, metaclass=QABCMeta):
    @abc.abstractmethod
    def func(self):
        pass

class MetaGLGraphics(type(QtCore.QObject), ABCMeta):
    pass

class GLGraphicsObject(QtWidgets.QGraphicsObject, AbstractGLGraphicsItem, metaclass=QABCMeta): 
    _qtBaseClass = QtWidgets.QGraphicsObject

GLGraphicsObject



import pyqtgraph 

unique_attr = set(dir(pyqtgraph.ViewBox)) - set(dir(pyqtgraph.GraphicsWidget))
class CustomViewBox(pyqtgraph.ViewBox):
    
    unique_attr = unique_attr
    
    def __getattr__(self, name):
        
        if not name in CustomViewBox.unique_attr:
            raise AttributeError(
                f"'{type(self).__name__}' object blocks access to inherited "
                f"GraphicsWidget attribute '{name}'"
            )
        
        # Fallback to normal attribute resolution
        return super().__getattr__(name)

from PySide6 import QtWidgets 
app = QtWidgets.QApplication()

vb = CustomViewBox()
vb.setSizePolicy  # Raises AttributeError if only in GraphicsWidget
import sys 
sys.exit()


#%%%

import abc
from PySide6.QtCore import QObject
from pyqtgraph.Qt import QT_LIB, QtCore, QtWidgets
from pyqtgraph.opengl import GLGraphicsItem


class QABCMeta(abc.ABCMeta, type(QObject)):
    

    def __new__(mcls, name, bases, ns, **kw):
        print(f"name: {name}")
        print(f"bases: {bases}")
        print(f'ns: {ns}')
        res = super.__new__.__qualname__
        print(f"res: {res}")


        cls = super().__new__(mcls, name, bases, ns, **kw)
        abc._abc_init(cls)
        return cls
    def __call__(cls, *args, **kw):
        if cls.__abstractmethods__:
            raise TypeError(f"Can't instantiate abstract class {cls.__name__} without an implementation for abstract methods {set(cls.__abstractmethods__)}")
        return super().__call__(*args, **kw)

class AbstractGLGraphicsItem(QtWidgets.QGraphicsItem, GLGraphicsItem.GLGraphicsItem, metaclass=QABCMeta):
    pass
class GLGraphicsObject(QtWidgets.QGraphicsObject, AbstractGLGraphicsItem):
    pass



#%%


class GrandParent1:
    def __init__(self, kwarg1=None):
        self.kwarg1 = kwarg1
        print(kwarg1)
        print("GrandParent1")
        super().__init__(kwarg1=kwarg1)
     

class GrandParent2:
    def __init__(self, kwarg1=None):
        self.kwarg1 = kwarg1
        print(kwarg1)
        print("GrandParent2")
        super().__init__()


class Parent1(GrandParent1, GrandParent2):
    def __init__(self, kwarg1=None):
        self.kwarg1 = kwarg1
        print("Parent1")
        #super().__init__()
        super().__init__(kwarg1=kwarg1)

class Parent2(GrandParent1):
    def __init__(self, kwarg1=None):
        self.kwarg1 = kwarg1
        print("Parent2")
        #super().__init__()
        super().__init__(kwarg1=kwarg1)

class Child1(Parent1, GrandParent1):
    def __init__(self, kwarg1=None):
        self.kwarg1 = kwarg1
        print("Child1")
        super().__init__(kwarg1=kwarg1)
        #super().__init__(kwarg1=kwarg1)
        


child = Child1(kwarg1=10)



#%%
from PySide6 import QtCore, QtGui, QtWidgets

from pyvol_terminal.gl_3D_graphing import meta
from pyqtgraph.opengl import GLGraphicsItem


class AbstractGLGraphicsItem(QtWidgets.QGraphicsItem, GLGraphicsItem.GLGraphicsItem, metaclass=meta.QABCMeta):
    
    def __init__(self, parent=None, **kwargs):
        print("\nGraphicsItem")
        print(f"\n{self}\n")
        QtWidgets.QGraphicsItem.__init__(self)

        self._viewBox=None
        self._blockUpdate=None
        
class GLGraphicsObject(QtWidgets.QGraphicsObject, AbstractGLGraphicsItem): 
    _qtBaseClass = QtWidgets.QGraphicsObject
    def __init__(self, parent=None):
        print(f"\nGraphicsObject")
        self.__inform_view_on_changes = True
        QtWidgets.QGraphicsObject.__init__(self, parent=parent)

class AbstractGLPlotItem(GLGraphicsObject, metaclass=meta.QABCMeta):
    sigPlotChanged = QtCore.Signal()
    
    def __init__(self, *args, view_box=None, **kwargs):
        self._viewBox=view_box        
        print("PlotItem")
        GLGraphicsObject.__init__(self, **kwargs)
        self.extra_opts={
                   "clipToView" : False
                   }
        self._boundsCache=[None]*3
        
        self.setProperty('xViewRangeWasChanged', False)
        self.setProperty('yViewRangeWasChanged', False)
        self.setProperty('zViewRangeWasChanged', False)
        self.setProperty('styleWasChanged', True)  # force initial update

plot_item = AbstractGLPlotItem()
"""

#%%%

from PySide6 import QtCore, QtGui, QtWidgets
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
from PySide6.QtWidgets import QGraphicsItem
import warnings
from pyqtgraph.Qt import QT_LIB
from pyvol_terminal.gl_3D_graphing.graphics_items.AbstractGLGraphicsItem import AbstractGLGraphicsItem

from pyqtgraph.opengl import GLGraphicsItem
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets, isQObjectAlive
from pyvol_terminal.gl_3D_graphing import meta

class AbstractGLGraphicsItem(GLGraphicsItem.GLGraphicsItem, QtWidgets.QGraphicsItem, metaclass=meta.QABCMeta):

    def __init__(self, parentItem=None, **kwargs):
        self._blockUpdate: bool
        self._viewBox = None
        self._viewWidget = None
        super().__init__()

        self._GLGraphicsItem__parent = None 
        self._GLGraphicsItem__children = list() 
        self._GLGraphicsItem__view = None 
        self._cachedView = None
        self.setParentItem(parentItem)


    @property
    def parent(self):
        """Getter for parent property"""
        return self._GLGraphicsItem__parent

    @parent.setter
    def parent(self, item):
        """Setter for parent property with all the logic"""
        if self._GLGraphicsItem__parent is not None:
            self._GLGraphicsItem__parent._GLGraphicsItem__children.remove(self)
        if item is not None:
            item._GLGraphicsItem__children.append(self)
            item.itemChange(item.GraphicsItemChange.ItemChildAddedChange, self)
            

        if self._GLGraphicsItem__view is not None:
            self._GLGraphicsItem__view.removeItem(self)

        self._GLGraphicsItem__parent = item
        self.itemChange(self.GraphicsItemChange.ItemParentHasChanged, item)
        self._GLGraphicsItem__view = None

    def setParentItem(self, item):
        self.parent = item


class CustomAbstractGLGraphicsItem(AbstractGLGraphicsItem):
    def __init__(self, parent=None):
        self.__inform_view_on_changes=True
        
        super().__init__()

    def boundingRect(self, *args, **kwargs):
        return

    def itemChange(self, change, value):
        ret = QtWidgets.QGraphicsItem.itemChange(self, change, value)
        print(f"\nPlotItem itemChange")
        print(f"change: {change}")
        print(f"value: {value}")
        if change in [self.GraphicsItemChange.ItemParentHasChanged, self.GraphicsItemChange.ItemSceneHasChanged]:
            print(True)
            if self.__class__.__dict__.get('parentChanged') is not None:
                # user's GraphicsObject subclass has a parentChanged() method
                warnings.warn(
                    "parentChanged() is deprecated and will be removed in the future. "
                    "Use changeParent() instead.",
                    DeprecationWarning, stacklevel=2
                )
                if QT_LIB == 'PySide6' and QtCore.__version_info__ == (6, 2, 2):
                    # workaround PySide6 6.2.2 issue https://bugreports.qt.io/browse/PYSIDE-1730
                    # note that the bug exists also in PySide6 6.2.2.1 / Qt 6.2.2
                    getattr(self.__class__, 'parentChanged')(self)
                else:
                    self.parentChanged()
            else:
                self.changeParent()
        try:
            inform_view_on_change = self.__inform_view_on_changes
        except AttributeError:
            # It's possible that the attribute was already collected when the itemChange happened
            # (if it was triggered during the gc of the object).
            pass
        else:
            if inform_view_on_change and change in [self.GraphicsItemChange.ItemPositionHasChanged, self.GraphicsItemChange.ItemTransformHasChanged]:
                self.informViewBoundsChanged()
        return ret

    
    def _internal_update(self, *args, **kwargs):...

graphics_itepaint = CustomAbstractGLGraphicsItem()
graphics_item2 = CustomAbstractGLGraphicsItem()

graphics_itepaint.setParentItem(graphics_item2)
import sys
sys.exit()


#%%


from abc import ABCMeta, ABC, abstractmethod

class GrandParent1(ABC):
    def __init__(self):
        super().__init__()

    def __init_subclass__(cls):
        print(cls)
    
    def paint(self):
        print("g")
        
    @abstractmethod
    def paintChild(self):...


class Parent1(GrandParent1):
    def __init__(self):
        super().__init__()
        
    def paint(self):
        super().paint()
        print("p")
        
class Parent2:
    def __init__(self):
        super().__init__()
        
    def paint(self):
        
        print("p")
        
    def abs(self):
        print("p abs")


class Child(Parent1, Parent2):
    pass
    def paintChild(self):
        print("c")


class ChildChild(Child):
    pass

    def paintChild(self):
        print("c")


c = Child()
c.paintChild()

#%%

class GrandParent1:
    def __init__(self, **kwargs):
        print("init GP1")
        #self.__x=None
        super().__init__(**kwargs)
        
    def paint(self):
        print(f"GBase1 paint")
        result = self.paintGL()
        return result

class GrandParent2:
    def __init__(self):
        print("init GP2")
        super().__init__()

class Parent1(GrandParent1):
    def __init__(self, **kwargs):
        print("init p1")
        self.__x=None
        super().__init__(**kwargs)

class Parent2(GrandParent2):
    def __init__(self, x):
        print("init p2")
        super().__init__()
        self.setData(x)
        
    def setData(self, x):
        print(f"Parent2 setData")
        self.__x=x



class Child(Parent1, Parent2):
    def __init__(self, x=None, **kwargs):
        print("Child")
        super().__init__(x=x, **kwargs)

    def paintGL2(self):
        print("Extend paintGL")
        Parent2.paint(self)


# use it like this:
e = Child(x=10)
#e.paint()



#%%

_x=2*np.arange(3)
import numpy as np
def func1(x=None):
    if not x is None:
        print("True func1")
        _x=x
    else:
        print("False func1")
        
        
def func2(x):
    if not x is None and not np.array_equal(x, _x):
        print("True func2")
    else:
        print("False func2")

#func2(None)
func2(2*np.arange(3))

_x=2*np.arange(3)

def func11(x=None):
    if not x is None:
        print("True func11")
        _x=x
    else:
        print("False func11")
        
        
def func22(x):
    if np.array_equal(x, _x):
        x=None
    func11(x=x)
#func22(None)
func22(2*np.arange(3))

